#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import re

import netius

from . import proxy_r

class DockerProxyServer(proxy_r.ReverseProxyServer):
    """
    Specialized reverse proxy server that handles many of the
    inner details of the reverse proxy server configuration
    server using environment variables information, making it
    easier for the automatic configuration and association of
    servers to the proxy server in a Docker environment.

    Although an optimal usage implies a Docker context the usage
    of the Docker proxy servers is not limited to such use cases.
    """

    def __init__(self, host_suffixes = [], *args, **kwargs):
        proxy_r.ReverseProxyServer.__init__(self, *args, **kwargs)
        self.load_config(host_suffixes = host_suffixes)
        self._build_docker()

    def on_serve(self):
        proxy_r.ReverseProxyServer.on_serve(self)
        if self.env: self.host_suffixes = self.get_env(
            "HOST_SUFFIXES",
            self.host_suffixes,
            cast = list
        )
        self._build_suffixes()
        self._build_redirect_ssl()

    def _build_docker(self):
        self._build_regex()
        self._build_hosts()
        self._build_alias()
        self._build_passwords()
        self._build_redirect()
        self._build_error_urls()
        self._build_redirect_ssl()

    def _build_regex(self, token = "$", sort = True):
        # retrieves the complete set of configuration values with the
        # regex suffix so that they are going to be used for the creation
        # of the regex rules (as expected)
        linked = netius.conf_suffix("_REGEX")

        # retrieves the complete set of names from the linked items and then
        # in case the sort flag is set sorts their values (proper order)
        names = netius.legacy.keys(linked)
        if sort: names.sort()

        # iterates over the complete set of linked regex values splitting
        # the values around the proper token and adding them to the regex
        for name in names:
            value = linked[name]
            value_s = value.split(token, 1)
            if not len(value_s) == 2: continue
            regex, target = value_s
            if not self._valid_url(target): continue
            rule = (re.compile(regex), target)
            self.regex.append(rule)

    def _build_hosts(self, alias = True):
        # tries to retrieve the complete set of configuration
        # values associated with the port suffix, this represents
        # the possible linked container addresses
        linked = netius.conf_suffix("_PORT")

        # iterates over the linked values, validating them and adding
        # them to the list of registered hosts
        for name, host in netius.legacy.iteritems(linked):
            # retrieves the name part of the configuration name
            # and converts it into lower cased value, note that
            # an extra dashed version is created, so that a proper
            # alias may be created for such naming
            base = name[:-5].lower()
            base_dash = base.replace("_", "-")

            # "builds" the name reference of the service and tries
            # to retrieve it from the configuration, it should exist
            # in case this port value represent a service
            name_ref = base.upper() + "_NAME"
            name_value = netius.conf(name_ref, None)
            if not name_value: continue

            # runs a series of validation on both the base and name
            # value to make sure that this value represents a valid
            # linked service/container
            # linked service/container (valid name reference found)
            if name.endswith("_ENV_PORT"): continue
            if not name.find("_ENV_") == -1: continue
            if base[-1].isdigit() and name_value[-1].isdigit(): continue

            # validates that the provided host is a valid URL value and
            # if that's not the case continues the loop (ignores)
            if not self._valid_url(host): continue

            # replaces the prefix of the reference (assumes HTTP) and
            # then adds the base value to the registered hosts
            host = host.replace("tcp://", "http://")
            host = str(host)
            self.hosts[base] = host

            # validates that the dashed version of the name is not the
            # same as the base one (at least one underscore) and if that's
            # not the case skips the current iteration
            if base == base_dash: continue

            # checks if the alias based registration is enabled and adds
            # the dashed version as an alias for such case or as an host
            # otherwise (static registration)
            if alias: self.alias[base_dash] = base
            else: self.hosts[base_dash] = host

    def _build_alias(self):
        linked = netius.conf_suffix("_ALIAS")
        for name, host in netius.legacy.iteritems(linked):
            base = name[:-6].lower()
            base_dash = base.replace("_", "-")
            self.alias[base] = host
            self.alias[base_dash] = host

    def _build_passwords(self):
        linked = netius.conf_suffix("_PASSWORD")
        for name, password in netius.legacy.iteritems(linked):
            base = name[:-9].lower()
            base_dash = base.replace("_", "-")
            simple_auth = netius.SimpleAuth(password = password)
            self.auth[base] = simple_auth
            self.auth[base_dash] = simple_auth

    def _build_redirect(self):
        linked = netius.conf_suffix("_REDIRECT")
        for name, host in netius.legacy.iteritems(linked):
            base = name[:-9].lower()
            base_dash = base.replace("_", "-")
            self.redirect[base] = host
            self.redirect[base_dash] = host

    def _build_error_urls(self):
        linked = netius.conf_suffix("_ERROR_URL")
        for name, error_url in netius.legacy.iteritems(linked):
            base = name[:-10].lower()
            base_dash = base.replace("_", "-")
            self.error_urls[base] = error_url
            self.error_urls[base_dash] = error_url

    def _build_redirect_ssl(self, alias = True):
        linked = netius.conf_suffix("_REDIRECT_SSL")
        for name, _force in netius.legacy.iteritems(linked):
            base = name[:-13].lower()
            base_dash = base.replace("_", "-")
            self.redirect[base] = (base, "https")
            self.redirect[base_dash] = (base_dash, "https")
            if not alias: continue
            for key, value in netius.legacy.iteritems(self.alias):
                is_match = value in (base, base_dash)
                if not is_match: continue
                self.redirect[key] = (key, "https")

    def _build_suffixes(self, alias = True, redirect = True):
        for host_suffix in self.host_suffixes:
            self.info("Registering %s host suffix" % host_suffix)
            for alias, value in netius.legacy.items(self.alias):
                fqn = alias + "." + str(host_suffix)
                self.alias[fqn] = value
            for name, value in netius.legacy.items(self.hosts):
                fqn = name + "." + str(host_suffix)
                if alias: self.alias[fqn] = name
                else: self.hosts[fqn] = value

    def _valid_url(self, value):
        """
        Verifies that the provided value is a valid URL
        value according to the default parsed.

        A valid URL should contain both a valid scheme and
        a not null and valid hostname.

        :type value: String
        :param value: The value that is going to be validated
        for URL compliance.
        :rtype: bool
        :return: If the provided value qualifies as a URL value.
        """

        value = str(value)
        result = netius.legacy.urlparse(value)
        if not result.scheme: return False
        if not result.hostname: return False
        return True

if __name__ == "__main__":
    server = DockerProxyServer()
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import ctypes

import netius

from . import base

class FsAdapter(base.BaseAdapter):

    def __init__(self, base_path = None):
        base.BaseAdapter.__init__(self)
        self.base_path = base_path or "fs.data"
        self.base_path = os.path.abspath(self.base_path)
        self.base_path = os.path.normpath(self.base_path)
        if not os.path.exists(self.base_path):
            os.makedirs(self.base_path)

    def set(self, value, owner = "nobody"):
        key = self.generate()
        owner_path = self._ensure(owner)
        file_path = os.path.join(self.base_path, key)
        link_path = os.path.join(owner_path, key)
        value = netius.legacy.bytes(value)
        file = open(file_path, "wb")
        try: file.write(value)
        finally: file.close()
        self._symlink(file_path, link_path)
        return key

    def get_file(self, key, mode = "rb"):
        file_path = os.path.join(self.base_path, key)
        file = open(file_path, mode)
        return file

    def delete(self, key, owner = "nobody"):
        owner_path = self._ensure(owner)
        file_path = os.path.join(self.base_path, key)
        link_path = os.path.join(owner_path, key)
        os.remove(file_path)
        os.remove(link_path)

    def size(self, key):
        file_path = os.path.join(self.base_path, key)
        return os.path.getsize(file_path)

    def count(self, owner = None):
        list = self.list(owner = owner)
        return len(list)

    def list(self, owner = None):
        path = self._path(owner = owner)
        exists = os.path.exists(path)
        files = os.listdir(path) if exists else []
        return files

    def _path(self, owner = None):
        if not owner: return self.base_path
        return os.path.join(self.base_path, owner)

    def _ensure(self, owner):
        owner_path = os.path.join(self.base_path, owner)
        if os.path.exists(owner_path): return owner_path
        os.makedirs(owner_path)
        return owner_path

    def _symlink(self, source, target):
        if os.name == "nt":
            symlink = ctypes.windll.kernel32.CreateSymbolicLinkW #@UndefinedVariable
            symlink.argtypes = (
                ctypes.c_wchar_p,
                ctypes.c_wchar_p,
                ctypes.c_uint32
            )
            symlink.restype = ctypes.c_ubyte
            flags = 1 if os.path.isdir(source) else 0
            result = symlink(target, source, flags)
            if result == 0: raise ctypes.WinError()
        else:
            os.symlink(source, target) #@UndefinedVariable



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import sys

import netius

from . import http
from . import http2

SERVER_SOFTWARE = netius.IDENTIFIER
""" The server software string that is going to identify the
current service that is running on the host, the values should
include both the name and the version of it """

COMPRESSED_LIMIT = 5242880
""" The default maximum size value for the sending of compressed
content, this should ensure proper resource usage avoiding extreme
high levels of resource usage for compression of large files """

class WSGIServer(http2.HTTP2Server):
    """
    Base class for the creation of a wsgi compliant server
    the server should be initialized with the "target" app
    object as reference and a mount point.
    """

    def __init__(
        self,
        app,
        mount = "",
        decode = True,
        compressed_limit = COMPRESSED_LIMIT,
        *args,
        **kwargs
    ):
        http2.HTTP2Server.__init__(self, *args, **kwargs)
        self.app = app
        self.mount = mount
        self.mount_l = len(mount)
        self.decode = decode
        self.compressed_limit = compressed_limit

    def on_connection_d(self, connection):
        http2.HTTP2Server.on_connection_d(self, connection)

        # tries to run the releasing operation on the current connection
        # so that the proper destruction of objects is performed avoiding
        # leaving any extra memory leak (would create problems)
        self._release(connection)

        # runs the extra release queue operation for the connection so that
        # the (possible) associated queue is properly release (no leaks)
        self._release_queue(connection)

    def on_serve(self):
        http2.HTTP2Server.on_serve(self)
        if self.env: self.compressed_limit = self.get_env(
            "COMPRESSED_LIMIT", self.compressed_limit, cast = int
        )
        self.info(
            "Starting WSGI server with %d bytes limit on compression ..." %\
            self.compressed_limit
        )

    def on_data_http(self, connection, parser):
        http2.HTTP2Server.on_data_http(self, connection, parser)

        # retrieves the path for the current request and then retrieves
        # the query string part for it also, after that computes the
        # path info value as the substring of the path without the mount
        path = parser.get_path(normalize = True)
        query = parser.get_query()
        path_info = path[self.mount_l:]

        # verifies if the path and query values should be encoded and if
        # that's the case the decoding process should unquote the received
        # path and then convert it into a valid string representation, this
        # is especially relevant for the python 3 infra-structure, this is
        # a tricky process but is required for the wsgi compliance
        if self.decode: path_info = self._decode(path_info)

        # retrieves a possible forwarded protocol value from the request
        # headers and calculates the appropriate (final scheme value)
        # taking the proxy value into account
        forwarded_protocol = parser.headers.get("x-forwarded-proto", None)
        scheme = "https" if connection.ssl else "http"
        scheme = forwarded_protocol if forwarded_protocol else scheme

        # initializes the environment map (structure) with all the cgi based
        # variables that should enable the application to handle the request
        # and respond to it in accordance
        environ = dict(
            REQUEST_METHOD = parser.method.upper(),
            SCRIPT_NAME = self.mount,
            PATH_INFO = path_info,
            QUERY_STRING = query,
            CONTENT_TYPE = parser.headers.get("content-type", ""),
            CONTENT_LENGTH = "" if parser.content_l == -1 else parser.content_l,
            SERVER_NAME = self.host,
            SERVER_PORT = str(self.port),
            SERVER_PROTOCOL = parser.version_s,
            SERVER_SOFTWARE = SERVER_SOFTWARE,
            REMOTE_ADDR = connection.address[0]
        )

        # updates the environment map with all the structures referring
        # to the wsgi specifications note that the message is retrieved
        # as a buffer to be able to handle the file specific operations
        environ["wsgi.version"] = (1, 0)
        environ["wsgi.url_scheme"] = scheme
        environ["wsgi.input"] = parser.get_message_b(copy = True)
        environ["wsgi.errors"] = sys.stderr
        environ["wsgi.multithread"] = False
        environ["wsgi.multiprocess"] = False
        environ["wsgi.run_once"] = False
        environ["wsgi.server_name"] = netius.NAME
        environ["wsgi.server_version"] = netius.VERSION

        # iterates over all the header values that have been received
        # to set them in the environment map to be used by the wsgi
        # infra-structure, not that their name is capitalized as defined
        # in the standard specification
        for key, value in netius.legacy.iteritems(parser.headers):
            key = "HTTP_" + key.replace("-", "_").upper()
            if isinstance(value, (list, tuple)): value = ";".join(value)
            environ[key] = value

        # verifies if the connection already has an iterator associated with
        # it, if that's the case the connection is already in use and the current
        # request processing must be delayed for future processing, this is
        # typically associated with http pipelining
        if hasattr(connection, "iterator") and connection.iterator:
            if not hasattr(connection, "queue"): connection.queue = []
            connection.queue.append(environ)
            return

        # calls the proper on environment callback so that the current request
        # is handled and processed (flush operation)
        self.on_environ(connection, environ)

    def on_environ(self, connection, environ):
        # method created as a clojure that handles the starting of
        # response as defined in the wsgi standards
        def start_response(status, headers):
            return self._start_response(connection, status, headers)

        # runs the app logic with the provided environment map and starts
        # response clojure and then iterates over the complete set of values
        # in the returned iterator to send the messages to the other end
        # note that the iterator and the environment map are set in the
        # connection for latter retrieval (required for processing/closing)
        sequence = self.app(environ, start_response)
        iterator = iter(sequence)
        connection.iterator = iterator
        connection.environ = environ

        # triggers the start of the connection iterator flushing operation
        # by calling the send part method for the current connection, this
        # should start reading data from the iterator and sending it to the
        # connection associated (recursive approach)
        self._send_part(connection)

    def _next_queue(self, connection):
        # verifies if the current connection already contains a reference to
        # the queue structure that handles the queueing/pipelining of requests
        # if it does not or the queue is empty returns immediately, as there's
        # nothing currently pending to be done/processed
        if not hasattr(connection, "queue"): return
        if not connection.queue: return

        # retrieves the current/first element in the connection queue to for
        # the processing and then runs the proper callback for the environ
        environ = connection.queue.pop(0)
        self.on_environ(connection, environ)

    def _start_response(self, connection, status, headers):
        # retrieves the parser object from the connection and uses
        # it to retrieve the string version of the http version
        parser = connection.parser
        version_s = parser.version_s

        # adds an extra space to the status line and then
        # splits it into the status message and the code
        status_c, status_m = (status + " ").split(" ", 1)
        status_c = int(status_c)
        status_m = status_m.strip()

        # converts the provided list of header tuples into a key
        # values based map so that it may be used more easily
        headers = dict(headers)

        # tries to retrieve the content length value from the headers
        # in case they exist and if the value of them is zero the plain
        # encoding is set in order to avoid extra problems while using
        # chunked encoding with zero length based messages
        length = headers.get("Content-Length", -1)
        length = int(length)
        length = 0 if status_c in (204, 304) else length
        if length == 0: connection.set_encoding(http.PLAIN_ENCODING)

        # verifies if the length value of the message payload overflow
        # the currently defined limit, if that's the case the connection
        # is set as uncompressed to avoid unnecessary encoding that would
        # consume a lot of resources (mostly processor)
        if length > self.compressed_limit: connection.set_uncompressed()

        # tries to determine if the accept ranges value is set and if
        # that's the case forces the uncompressed encoding to avoid possible
        # range missmatch due to re-encoding of the content
        ranges = headers.get("Accept-Ranges", None)
        if ranges == "bytes": connection.set_uncompressed()

        # determines if the content range header is set, meaning that
        # a partial chunk value is being sent if that's the case the
        # uncompressed encoding is forced to avoid re-encoding issues
        content_range = headers.get("Content-Range", None)
        if content_range: connection.set_uncompressed()

        # verifies if the current connection is using a chunked based
        # stream as this will affect some of the decisions that are
        # going to be taken as part of response header creation
        is_chunked = connection.is_chunked()

        # checks if the provided headers map contains the definition
        # of the content length in case it does not unsets the keep
        # alive setting in the parser because the keep alive setting
        # requires the content length to be defined or the target
        # encoding type to be chunked
        has_length = not length == -1
        if not has_length: parser.keep_alive = is_chunked

        # applies the base (static) headers to the headers map and then
        # applies the parser based values to the headers map, these
        # values should be dynamic and based in the current state
        # finally applies the connection related headers to the current
        # map of headers so that the proper values are filtered and added
        self._apply_base(headers)
        self._apply_parser(parser, headers)
        self._apply_connection(connection, headers)

        # runs the send header operation on the connection, this operation
        # should serialize the various headers and send them through the
        # current connection according to the currently associated protocol
        connection.send_header(
            headers = headers,
            version = version_s,
            code = status_c,
            code_s = status_m
        )

    def _send_part(self, connection):
        # invalidates the data object to the original unset value,
        # this is considered the default value
        data = None

        # extracts both the iterator from the connection object so that
        # it may be used for the current set of operations, if there's
        # no iterator associated with the connection the current iteration
        # is marked as final (should be flushed immediately)
        iterator = connection.iterator
        is_final = False if iterator else True

        # tries to retrieve data from the current iterator and in
        # case the stop iteration is received sets the is final flag
        # so that no more data is sent through the connection and
        # releases the iterator from the connection
        try: data = next(iterator) if iterator else None
        except StopIteration as exception:
            # tries to extract possible data coming from the exception
            # return value and sets the is final flag otherwise
            if exception.args: data = exception.args[0]
            else: is_final = True

            # releases the iterator from the connection as it's no longer
            # considered to be valid for the current connection context
            self._release_iterator(connection)

        # verifies if the current value in iteration is a future element
        # and if that's the case creates the proper callback to be used
        # for the handling on the end of the iteration
        is_future = netius.is_future(data)
        if is_future:
            def on_partial(future, value):
                if not value: return
                if not self.is_main(): return self.delay_s(
                    lambda: on_partial(future, value)
                )
                connection.send_part(value, final = False)

            def on_done(future):
                # in case the current threads is not the main one (running using
                # thread pool) delays the current callback to be called upon
                # the next main event loop ticks
                if not self.is_main(): return self.delay_s(
                    lambda: on_done(future)
                )

                # unsets the future from the connection as it has been
                # completely processed, not going to be used anymore
                connection.future = None

                # verifies if the future has been canceled in the mean time
                # (error situation) and if that's the case closes the connection
                # and raises an exception so that the event loop handles it gracefully
                if future.cancelled():
                    connection.close()
                    raise RuntimeError("Problem handling future, canceled")

                # verifies if the future contains an exception value in it (error
                # situation) and if that's the case closes the connection and raises
                # the exception so that the event loop handles it gracefully
                if future.exception():
                    connection.close()
                    raise future.exception()

                # otherwise runs the send part operation on the next tick so
                # that it gets handled as fast as possible, this should continue
                # the iteration on the overall async generator
                else: self.delay(
                    lambda: self._send_part(connection),
                    immediately = True
                )

            def on_ready():
                return connection.wready

            def on_closed():
                return connection.is_closed()

            # registers the various callbacks that are going to be
            # handling the various operations on the future element
            # so that the iterator may go further into the next
            # elements, after the processing of this future
            data.add_partial_callback(on_partial)
            data.add_done_callback(on_done)
            data.add_ready_callback(on_ready)
            data.add_closed_callback(on_closed)

            # sets the current future object in the connection so that
            # it may be latter usage for possible release
            connection.future = data

            # returns the control flow immediately as there's nothing
            # left to be done for a future operation at the moment
            return

        # ensures that the provided data is a byte sequence as expected
        # by the underlying server infra-structure
        if data: data = netius.legacy.bytes(data)

        # in case the final flag is set runs the flush operation in the
        # connection setting the proper callback method for it so that
        # the connection state is defined in the proper way (closed or
        # kept untouched) otherwise sends the retrieved data setting the
        # callback to the current method so that more that is sent
        if is_final: connection.flush_s(callback = self._final)
        else: connection.send_part(
            data,
            final = False,
            callback = self._send_part
        )

    def _final(self, connection):
        # retrieves the parser of the current connection and then determines
        # if the current connection is meant to be kept alive
        parser = connection.parser
        keep_alive = parser.keep_alive

        # in case the connection is not meant to be kept alive must
        # must call the proper underlying close operation (expected)
        if not keep_alive: self._close(connection); return

        # the map of environment must be destroyed properly, avoiding
        # any possible memory leak for the current handling and then the
        # queue of pipelined requests must be flushed/processed, this
        # allows the connection to be re-used for new/pending requests
        self._release(connection)
        self._next_queue(connection)

    def _close(self, connection):
        connection.close(flush = True)

    def _release(self, connection):
        self._release_future(connection)
        self._release_iterator(connection)
        self._release_environ(connection)
        self._release_parser(connection)

    def _release_future(self, connection):
        # verifies if there's a future associated/running under the
        # current connection, if that's not the case returns immediately
        future = hasattr(connection, "future") and connection.future
        if not future: return

        # runs the cancel operation on the future, note that this
        # operation is only performed in case the future is still
        # under the running state (normal operation)
        future.cancel()

        # unsets the future from the connection as it's no longer
        # associated with it (as it's considered closed)
        connection.future = None

    def _release_iterator(self, connection):
        # verifies if there's an iterator object currently defined
        # in the connection so that it may be close in case that's
        # required, this is mandatory to avoid any memory leak
        iterator = hasattr(connection, "iterator") and connection.iterator
        if not iterator: return

        # verifies if the close attributes is defined in the iterator
        # and if that's the case calls the close method in order to
        # avoid any memory leak caused by the generator
        has_close = hasattr(iterator, "close")
        if has_close: iterator.close()

        # unsets the iterator attribute in the connection object so that
        # it may no longer be used by any chunk of logic code
        connection.iterator = None

    def _release_environ(self, connection):
        # tries to retrieve the map of environment for the current
        # connection and in case it does not exists returns immediately
        environ = hasattr(connection, "environ") and connection.environ
        if not environ: return

        # retrieves the input stream (buffer) and closes it as there's
        # not going to be any further operation in it (avoids leak)
        input = environ["wsgi.input"]
        input.close()

        # removes the complete set of key to value associations in the
        # map and unsets the environ value in the current connection
        environ.clear()
        connection.environ = None

    def _release_parser(self, connection):
        # closes the current file objects in the parser, note that the
        # parser still remains active, this operation only clears the
        # current memory structures associated with the parser
        connection.parser.close()

    def _release_queue(self, connection):
        # tries to retrieve a possible defined queue for the provided
        # connection in case it does not exist returns immediately as
        # there's no queue element to be release/cleared
        queue = hasattr(connection, "queue") and connection.queue
        if not queue: return

        # iterates over the complete set of queue elements (environ
        # based maps) to clear their elements properly
        for environ in queue:
            # retrieves the reference to the input buffer file and
            # closes it (avoiding possible extra accesses and leaks)
            input = environ["wsgi.input"]
            input.close()

            # empties the map key references so that no more access
            # to the map is possible (avoids leaks)
            environ.clear()

        # removes the complete set of elements from the queue while
        # maintaining the original list reference/instance
        del queue[:]

    def _decode(self, value):
        """
        Decodes the provided quoted value, normalizing it according
        to both the pep 333 and the pep 333.

        Note that python 3 enforces the encapsulation of the string
        value around a latin 1 encoded unicode string.

        :type value: String
        :param value: The quoted value that should be normalized and
        decoded according to the wsgi 1.0/1.0.1 specifications.
        :rtype: String
        :return: The normalized version of the provided quoted value
        that is ready to be provided as part of the environ map.
        :see: http://python.org/dev/peps/pep-3333
        """

        value = netius.legacy.unquote(value)
        is_unicode = netius.legacy.is_unicode(value)
        value = value.encode("utf-8") if is_unicode else value
        value = netius.legacy.str(value)
        return value

if __name__ == "__main__":
    import logging

    def app(environ, start_response):
        status = "200 OK"
        contents = "Hello World"
        content_l = len(contents)
        headers = (
            ("Content-Length", content_l),
            ("Content-type", "text/plain"),
            ("Connection", "keep-alive")
        )
        start_response(status, headers)
        yield contents

    server = WSGIServer(app = app, level = logging.INFO)
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius.common

from . import base

def generate(domain, suffix = None, number_bits = 1024):
    number_bits = int(number_bits)
    result = netius.common.dkim_generate(domain, suffix = suffix, number_bits = number_bits)
    print(result["dns_txt"])
    print(result["private_pem"])

def sign(email_path, key_path, selector, domain):
    file = open(email_path, "rb")
    try: contents = file.read()
    finally: file.close()

    contents = contents.lstrip()
    private_key = netius.common.open_private_key(key_path)
    signature = netius.common.dkim_sign(contents, selector, domain, private_key)

    file = open(email_path, "wb")
    try: file.write(signature); file.write(contents)
    finally: file.close()

if __name__ == "__main__":
    base.sh_call(globals(), locals())
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import base64
import datetime

import email.parser

import netius.common

from . import dns

HELO_STATE = 1

EHLO_STATE = 2

CAPA_STATE = 3

STLS_STATE = 4

UPGRADE_STATE = 5

AUTH_STATE = 6

USERNAME_STATE = 7

PASSWORD_STATE = 8

FROM_STATE = 9

TO_STATE = 10

DATA_STATE = 11

CONTENTS_STATE = 12

QUIT_STATE = 13

FINAL_STATE = 14

class SMTPConnection(netius.Connection):

    AUTH_METHODS = (
        "plain",
        "login"
    )
    """ The sequence that defined the multiple allowed
    methods for this SMTP protocol implementation """

    def __init__(self, host = "smtp.localhost", *args, **kwargs):
        netius.Connection.__init__(self, *args, **kwargs)
        self.parser = None
        self.host = host
        self.froms = None
        self.tos = None
        self.contents = None
        self.username = None
        self.password = None
        self.expected = None
        self.to_index = 0
        self.state = HELO_STATE
        self.next = None
        self.sindex = 0
        self.sequence = ()
        self.capabilities = ()
        self.messages = []

    def open(self, *args, **kwargs):
        netius.Connection.open(self, *args, **kwargs)
        if not self.is_open(): return
        self.parser = netius.common.SMTPParser(self)
        self.parser.bind("on_line", self.on_line)
        self.build()

    def close(self, *args, **kwargs):
        netius.Connection.close(self, *args, **kwargs)
        if not self.is_closed(): return
        if self.parser: self.parser.destroy()
        self.destroy()

    def build(self):
        """
        Builds the initial set of states ordered according to
        their internal integer definitions, this method provides
        a fast and scalable way of parsing data.
        """

        netius.Connection.build(self)
        self.states = (
            self.helo_t,
            self.ehlo_t,
            self.capa_t,
            self.stls_t,
            self.upgrade_t,
            self.auth_t,
            self.username_t,
            self.password_t,
            self.mail_t,
            self.rcpt_t,
            self.data_t,
            self.contents_t,
            self.quit_t,
            self.close_t
        )
        self.state_l = len(self.states)

    def destroy(self):
        """
        Destroys the current structure for the stats meaning that
        it's restored to the original values, this method should only
        be called on situation where no more client usage is required.
        """

        netius.Connection.destroy(self)
        self.states = ()
        self.state_l = 0

    def set_smtp(self, froms, tos, contents, username = None, password = None):
        self.froms = froms
        self.tos = tos
        self.contents = contents
        self.username = username
        self.password = password

    def set_sequence(self, sequence, safe = True):
        if safe and self.sequence == sequence: return
        self.sindex = 0
        self.sequence = sequence
        self.state = sequence[0]

    def set_message_seq(self, ehlo = True):
        sequence = (
            EHLO_STATE if ehlo else HELO_STATE,
            CAPA_STATE,
            AUTH_STATE,
            USERNAME_STATE,
            PASSWORD_STATE,
            FROM_STATE,
            TO_STATE,
            DATA_STATE,
            CONTENTS_STATE,
            QUIT_STATE,
            FINAL_STATE
        )
        self.set_sequence(sequence)

    def set_message_stls_seq(self, ehlo = True):
        sequence = (
            EHLO_STATE if ehlo else HELO_STATE,
            CAPA_STATE,
            STLS_STATE,
            UPGRADE_STATE,
            EHLO_STATE if ehlo else HELO_STATE,
            CAPA_STATE,
            AUTH_STATE,
            USERNAME_STATE,
            PASSWORD_STATE,
            FROM_STATE,
            TO_STATE,
            DATA_STATE,
            CONTENTS_STATE,
            QUIT_STATE,
            FINAL_STATE
        )
        self.set_sequence(sequence)

    def set_capabilities(self, capabilities, force = True):
        if not force and self.capabilities: return
        capabilities = [value.strip().lower() for value in capabilities]
        self.capabilities = capabilities

    def next_sequence(self):
        self.sindex += 1
        self.sindex = self.sequence.index(self.next) if self.next else self.sindex
        self.state = self.next if self.next else self.sequence[self.sindex]
        self.next = None

    def parse(self, data):
        return self.parser.parse(data)

    def send_smtp(self, code, message = "", delay = True, callback = None):
        base = "%s %s" % (code, message)
        data = base + "\r\n"
        count = self.send(data, delay = delay, callback = callback)
        self.owner.debug(base)
        return count

    def on_line(self, code, message, is_final = True):
        # creates the base string from the provided code value and the
        # message associated with it, then logs the values into the
        # current debug logger support (for traceability)
        base = "%s %s" % (code, message)
        self.owner.debug(base)

        # adds the message part of the line to the buffer that holds the
        # various messages "pending" for the current response, these values
        # may latter be used for the processing of the response
        self.messages.append(message)

        # in case the currently parsed line is not a final one must return
        # immediately to continue the processing of information for the
        # current response, the various message should be accumulated under
        # the message buffer to avoid any problem
        if not is_final: return

        # runs the code based assertion so that if there's an expected
        # value set for the current connection it's correctly validated
        self.assert_c(code)

        # verifies that the current state valid represents a valid state
        # according to the ones that have "generate" handling methods, otherwise
        # raises a parser error indicating the problem
        if self.state > self.state_l:
            raise netius.ParserError("Invalid state")

        # runs the calling of the next state based method according to the
        # currently defined state, this is the increments in calling
        self.call()

        # erases the message buffer as it's not longer going to be used in
        # the handling (as it is finished) nothing remaining
        del self.messages[:]

    def call(self):
        # tries to retrieve the method for the current state in iteration
        # and then calls the retrieve method with no arguments (handler method)
        method = self.states[self.state - 1]
        method()

    def skip(self):
        self.next_sequence()
        self.call()

    def helo_t(self):
        self.helo(self.host)
        self.next_sequence()

    def ehlo_t(self):
        self.ehlo(self.host)
        self.next_sequence()

    def capa_t(self):
        capabilities = self.messages[1:]
        self.set_capabilities(capabilities)
        if "starttls" in self.capabilities:
            self.set_message_stls_seq()
        self.next_sequence()
        self.call()

    def stls_t(self):
        self.starttls()
        self.next_sequence()

    def upgrade_t(self):
        def callback(connection):
            connection.upgrade(server = False)
        self.next_sequence()
        callback(self)

    def auth_t(self):
        is_valid = self.username and self.password
        if not is_valid:
            self.next = FROM_STATE
            self.skip()
            return

        method = self.best_auth()
        self.auth(self.username, self.password, method = method)
        self.next_sequence()

    def username_t(self):
        self.login_username(self.username)
        self.next_sequence()

    def password_t(self):
        self.login_password(self.password)
        self.next_sequence()

    def mail_t(self):
        self.mail(self.froms[0])
        self.next_sequence()

    def rcpt_t(self):
        is_final = self.to_index == len(self.tos) - 1
        self.rcpt(self.tos[self.to_index])
        self.to_index += 1
        if is_final: self.next_sequence()

    def data_t(self):
        self.data()
        self.next_sequence()

    def contents_t(self):
        self.assert_s(CONTENTS_STATE)
        self.send(self.contents)
        self.send("\r\n.\r\n")
        self.set_expected(250)
        self.next_sequence()

    def quit_t(self):
        self.quit()
        self.next_sequence()

    def close_t(self):
        self.close(flush = True)

    def pass_t(self):
        pass

    def helo(self, host):
        self.assert_s(HELO_STATE)
        message = host
        self.send_smtp("helo", message)
        self.set_expected(250)

    def ehlo(self, host):
        self.assert_s(EHLO_STATE)
        message = host
        self.send_smtp("ehlo", message)
        self.set_expected(250)

    def starttls(self):
        self.assert_s(STLS_STATE)
        self.send_smtp("starttls")
        self.set_expected(220)

    def auth(self, username, password, method = "plain"):
        self.assert_s(AUTH_STATE)
        method_name = "auth_%s" % method
        has_method = hasattr(self, method_name)
        if not has_method: raise netius.NotImplemented("Method not implemented")
        method = getattr(self, method_name)
        method(username, password)

    def auth_plain(self, username, password):
        auth_value = "\0%s\0%s" % (username, password)
        auth_value = netius.legacy.bytes(auth_value)
        auth_value = base64.b64encode(auth_value)
        auth_value = netius.legacy.str(auth_value)
        message = "%s %s" % ("plain", auth_value)
        self.send_smtp("auth", message)
        self.set_expected(235)
        self.next = FROM_STATE

    def auth_login(self, username, password):
        message = "login"
        self.send_smtp("auth", message)
        self.set_expected(334)

    def login_username(self, username):
        username = netius.legacy.bytes(username)
        username = base64.b64encode(username)
        username = netius.legacy.str(username)
        self.send_smtp(username)
        self.set_expected(334)

    def login_password(self, password):
        password = netius.legacy.bytes(password)
        password = base64.b64encode(password)
        password = netius.legacy.str(password)
        self.send_smtp(password)
        self.set_expected(235)

    def mail(self, value):
        self.assert_s(FROM_STATE)
        message = "FROM:<%s>" % value
        self.send_smtp("mail", message)
        self.set_expected(250)

    def rcpt(self, value):
        self.assert_s(TO_STATE)
        message = "TO:<%s>" % value
        self.send_smtp("rcpt", message)
        self.set_expected(250)

    def data(self):
        self.assert_s(DATA_STATE)
        self.send_smtp("data")
        self.set_expected(354)

    def quit(self):
        self.assert_s(QUIT_STATE)
        self.send_smtp("quit")
        self.set_expected(221)

    def set_expected(self, expected):
        self.expected = expected

    def assert_c(self, code):
        if not self.expected: return
        expected = self.expected
        code_i = int(code)
        self.expected = None
        valid = expected == code_i
        if valid: return
        raise netius.ParserError(
            "Invalid response code expected '%d' received '%d'" %
            (expected, code_i)
        )

    def assert_s(self, expected):
        if self.state == expected: return
        raise netius.ParserError("Invalid state")

    def best_auth(self):
        cls = self.__class__
        methods = []
        for capability in self.capabilities:
            is_auth = capability.startswith("auth ")
            if not is_auth: continue
            parts = capability.split(" ")
            parts = [part.strip() for part in parts]
            methods.extend(parts[1:])
        usable = [method for method in methods if method in cls.AUTH_METHODS]
        return usable[0] if usable else "plain"

class SMTPClient(netius.StreamClient):

    def __init__(
        self,
        host = None,
        auto_close = False,
        *args,
        **kwargs
    ):
        netius.StreamClient.__init__(self, *args, **kwargs)
        self.host = host if host else "[" + netius.common.host() + "]"
        self.auto_close = auto_close

    @classmethod
    def message_s(
        cls,
        froms,
        tos,
        contents,
        daemon = True,
        host = None,
        mark = True,
        callback = None
    ):
        smtp_client = cls.get_client_s(thread = True, daemon = daemon, host = host)
        smtp_client.message(
            froms,
            tos,
            contents,
            mark = mark,
            callback = callback
        )

    def message(
        self,
        froms,
        tos,
        contents,
        host = None,
        port = 25,
        username = None,
        password = None,
        ehlo = True,
        stls = False,
        mark = True,
        ensure_loop = True,
        callback = None
    ):
        # in case the mark flag is set the contents data is modified
        # and "marked" with the pre-defined header values of the client
        # this should provide some extra information on the agent
        if mark: contents = self.mark(contents)

        # creates the method that is able to generate handler for a
        # certain sequence of to based (email) addresses
        def build_handler(tos, domain = None, tos_map = None):

            def on_close(connection = None):
                # verifies if the current handler has been build with a
                # domain based clojure and if that's the case removes the
                # reference of it from the map of tos, then verifies if the
                # map is still valid and if that's the case returns and this
                # is not considered the last remaining SMTP session for the
                # current send operation (still some open)
                if domain: del tos_map[domain]
                if tos_map: return

                # verifies if the callback method is defined and if that's
                # the case calls the callback indicating the end of the send
                # operation (note that this may represent multiple SMTP sessions)
                callback and callback(self)

            def handler(response = None):
                # in case the provided response value is invalid returns
                # immediately, as this should represent a resolution error,
                # this is only done in case the host is also not defined
                # as for such situations an address is not retrievable
                if response == None and host == None: return

                # in case there's a valid response provided must parse it
                # to try to "recover" the final address that is going to be
                # used in the establishment of the SMTP connection
                if response:
                    # in case there are no answers present in the response
                    # of the dns resolution an exception must be raised, note
                    # that the on close handler is called so that proper
                    # fallback for this connections is handled
                    if not response.answers:
                        on_close()
                        if self.auto_close: self.close()
                        raise netius.NetiusError(
                            "Not possible to resolve MX for '%s'" % domain
                        )

                    # retrieves the first answer (probably the most accurate)
                    # and then unpacks it until the mx address is retrieved
                    first = response.answers[0]
                    extra = first[4]
                    address = extra[1]

                # otherwise the host should have been provided and as such the
                # address value is set with the provided host
                else: address = host

                # sets the proper address (host) and port values that are
                # going to be used to establish the connection, notice that
                # in case the values provided as parameter to the message
                # method are valid they are used instead of the "resolved"
                _host = host or address
                _port = port or 25

                # prints a debug message about the connection that is now
                # going to be established (helps with debugging purposes)
                self.debug("Establishing SMTP connection on %s:%d ..." % (_host, _port))

                # establishes the connection to the target host and port
                # and using the provided key and certificate files and then
                # sets the SMTP information in the current connection, after
                # the connections is completed the SMTP session should start
                connection = self.connect(_host, _port)
                if stls: connection.set_message_stls_seq(ehlo = ehlo)
                else: connection.set_message_seq(ehlo = ehlo)
                connection.set_smtp(
                    froms,
                    tos,
                    contents,
                    username = username,
                    password = password
                )
                connection.bind("close", on_close)
                return connection

            # returns the clojure bound handler method, ready
            # to be used under a specific context
            return handler

        # in case the host address has been provided by argument the
        # handler method is called immediately to trigger the processing
        # of the SMTP connection using the current host and port
        if host:
            handler = build_handler(tos)
            connection = handler()
            return connection

        # ensures that the proper main loop is started so that the current
        # SMTP client does not become orphan as no connection has been
        # established as of this moment (as expected) and the dns client
        # is going to be run as a daemon (avoids process exit)
        if ensure_loop: self.ensure_loop()

        # creates the map that is going to be used to associate each of
        # the domains with the proper to (email) addresses, this is going
        # to allow aggregated based SMTP sessions (performance wise)
        tos_map = dict()
        for to in tos:
            _name, domain = to.split("@", 1)
            _tos = tos_map.get(domain, [])
            _tos.append(to)
            tos_map[domain] = _tos

        # iterates over the complete set of domain and associated
        # to addresses list for each of them to run the mx based
        # query operation and then start the SMTP session
        for domain, tos in netius.legacy.items(tos_map):
            # creates a new handler method bound to the to addresses
            # associated with the current domain in iteration
            handler = build_handler(tos, domain = domain, tos_map = tos_map)

            # prints a small debug message about the resolution of the
            # domain for the current message (debugging purposes)
            self.debug("Resolving MX domain for'%s' ..." % domain)

            # runs the dns query to be able to retrieve the proper
            # mail exchange host for the target email address and then
            # sets the proper callback for sending
            dns.DNSClient.query_s(domain, type = "mx", callback = handler)

    def on_connect(self, connection):
        netius.StreamClient.on_connect(self, connection)

    def on_upgrade(self, connection):
        netius.StreamClient.on_upgrade(self, connection)
        connection.call()

    def on_data(self, connection, data):
        netius.StreamClient.on_data(self, connection, data)
        connection.parse(data)

    def on_connection_d(self, connection):
        netius.StreamClient.on_connection_d(self, connection)
        if not self.auto_close: return
        if self.connections: return
        self.close()

    def build_connection(self, socket, address, ssl = False):
        return SMTPConnection(
            owner = self,
            socket = socket,
            address = address,
            ssl = ssl,
            host = self.host
        )

    def mark(self, contents):
        parser = email.parser.Parser()
        message = parser.parsestr(contents)
        self.date(message)
        self.user_agent(message)
        return message.as_string()

    def date(self, message):
        date = message.get("Date", None)
        if date: return
        date_time = datetime.datetime.utcnow()
        message["Date"] = date_time.strftime("%a, %d %b %Y %H:%M:%S +0000")

    def user_agent(self, message):
        user_agent = message.get("User-Agent", None)
        if user_agent: return
        message["User-Agent"] = netius.IDENTIFIER

if __name__ == "__main__":
    import email.mime.text

    sender = netius.conf("SMTP_SENDER", "hello@bemisc.com")
    receiver = netius.conf("SMTP_RECEIVER", "hello@bemisc.com")
    host = netius.conf("SMTP_HOST", None)
    port = netius.conf("SMTP_PORT", 25, cast = int)
    username = netius.conf("SMTP_USER", None)
    password = netius.conf("SMTP_PASSWORD", None)
    stls = netius.conf("SMTP_STARTTLS", False, cast = bool)

    mime = email.mime.text.MIMEText("Hello World")
    mime["Subject"] = "Hello World"
    mime["From"] = sender
    mime["To"] = receiver
    contents = mime.as_string()

    client = SMTPClient(auto_close = True)
    client.message(
        [sender],
        [receiver],
        contents,
        host = host,
        port = port,
        username = username,
        password = password,
        stls = stls
    )
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2017 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2017 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import errors
from . import observer

class Transport(observer.Observable):
    """
    Decorator class to be used to add the functionality of a
    transport layer using a simplified and standard API.

    Allows adding the functionality to an internal netius
    connection (or equivalent) object.

    This approach is heavily influenced by the design of the
    asyncio Python infra-structure and should provide a mostly
    compatible interface.
    """

    def __init__(self, loop, connection, open = True):
        self._loop = loop
        self._connection = connection
        self._protocol = None
        self._extra_dict = None
        self._exhausted = False
        if open: self.open()

    def open(self):
        self.set_handlers()
        self.set_write_buffer_limits()
        self.set_extra_dict()

    def close(self):
        # in case the current transport is already closed or in the
        # process of closing returns immediately (avoids duplication)
        if self.is_closing(): return

        # in case there's a connection object set schedules its closing
        # otherwise unsets the protocol object immediately
        if self._connection: self._connection.close(flush = True)
        else: self._protocol = None

        # removes the reference to the underlying connection object
        # and unsets the exhausted flag (reset of values)
        self._connection = None
        self._exhausted = False

    def abort(self):
        # in case the current transport is already closed or in the
        # process of closing returns immediately (avoids duplication)
        if self.is_closing(): return

        # in case there's a connection set runs the (forced) close
        # operation so that no more interaction exists otherwise
        # unsets the protocol (notice that if the connection exists
        # the close operation will trigger the close protocol callback)
        if self._connection: self._connection.close()
        else: self._protocol = None

        # unsets the connection object (as it's no longer eligible
        # to be used) and unsets the current transport for exhausted
        # as it's considered to be the default/initials state
        self._connection = None
        self._exhausted = False

    def write(self, data):
        # verifies if the current connection is closing or in the process
        # of closing and if that's the case returns immediately (graceful)
        if self.is_closing(): return

        # runs the send operation on the underlying (and concrete)
        # connection object, notice that the delay flag is unset so
        # that the send flushing operation runs immediately (to provide
        # behaviour level compatibility with the asyncio library)
        self._connection.send(data, delay = False)

    def sendto(self, data, addr = None):
        # verifies if the current connection is closing or in the process
        # of closing and if that's the case returns immediately (graceful)
        if self.is_closing(): return

        # runs the send operation on the underlying (and concrete)
        # connection object, notice that the delay flag is unset so
        # that the send flushing operation runs immediately (to provide
        # behaviour level compatibility with the asyncio library)
        self._connection.send(data, address = addr, delay = False)

    def get_extra_info(self, name, default = None):
        callable = self._extra_dict.get(name, None)
        if callable: return callable()
        else: return default

    def get_write_buffer_size(self):
        return self._connection.pending_s

    def get_write_buffer_limits(self):
        return (
            self._connection.min_pending,
            self._connection.max_pending
        )

    def set_handlers(self):
        self._connection.bind("pend", self._buffer_touched)
        self._connection.bind("unpend", self._buffer_touched)

    def set_write_buffer_limits(self, high = None, low = None):
        """
        Sets the write buffer limits in the underlying connection
        object using the provided values.

        If the only one of the values is provided the other one is
        going to be calculated using an hardcoded ratio value.

        :type high: int
        :param high: The maximum number of bytes that can be set
        waiting in the connection internal buffer waiting to be sent
        before the connection becomes exhausted (sending blocked).
        :type low: int
        :param low: The maximum number of bytes waiting in the buffer
        before the connection send buffer is unblocked.
        """

        if high is None:
            if low == None: high = 65536
            else: high = 4 * low
        if low == None: low = high // 4
        if not high >= low >= 0:
            raise errors.RuntimeError("High must be larger than low")

        self._connection.max_pending = high
        self._connection.min_pending = low

    def set_extra_dict(self):
        self._extra_dict = dict(
            socket = lambda: self._connection.socket,
            peername = lambda: self._connection.socket.getpeername(),
            sockname = lambda: self._connection.socket.getsockname(),
            compression = lambda: self._connection.socket.compression(),
            cipher = lambda: self._connection.socket.cipher(),
            peercert = lambda: self._connection.socket.getpeercert(),
            sslcontext = lambda: self._connection.socket.context,
            ssl_object = lambda: self._connection.socket
        )

    def get_protocol(self):
        return self._protocol

    def set_protocol(self, protocol):
        self._set_protocol(protocol, mark = False)

    def is_closing(self):
        """
        Determines if the current transport/connection is closed
        or is in the process of being closed.

        A transport is considered closed in case the underlying
        connection object is not set or is closed.

        :rtype: bool
        :return: If the current transport is closed or in the process
        of being closed, if the connection is not set in the transport
        it's also considered to be closing.
        """

        if not self._connection: return True
        if self._connection.is_closed(): return True
        return False

    def _on_data(self, connection, data):
        pass

    def _on_close(self, connection):
        self._cleanup()

    def _set_compat(self, protocol):
        self._set_binds()
        self._set_protocol(protocol)

    def _set_binds(self):
        self._connection.bind("data", self._on_data)
        self._connection.bind("close", self._on_close)

    def _set_protocol(self, protocol, mark = True):
        self._protocol = protocol
        if mark: self._protocol.connection_made(self)

    def _buffer_touched(self, connection):
        self._handle_flow()

    def _handle_flow(self):
        if not self._connection: return

        if self._exhausted:
            is_restored = self._connection.is_restored()
            if not is_restored: return
            self._exhausted = False
            self._protocol.resume_writing()
        else:
            is_exhausted = self._connection.is_exhausted()
            if not is_exhausted: return
            self._exhausted = True
            self._protocol.pause_writing()

    def _cleanup(self):
        # schedules the connection lost calling on on the
        # next available tick so that the associated transport
        # gets notified about the closing of the connection
        self._call_soon(self._call_connection_lost, None)

        # unsets the loop from the current transport as it's
        # not safe to interact with the owner loop anymore
        self._loop = None

        # unsets the connection from the current instance and
        # normalizes the value of the exhausted flag
        self._connection = None
        self._exhausted = False

        # unset the currently set extra dictionary that defines
        # a series of lambda functions to the current instance
        self._extra_dict = None

        # runs the destroy operation that will clean the
        # most basic structures associated with this object
        self.destroy()

    def _call_connection_lost(self, context):
        # verifies if there's a protocol instance currently
        # set and if that's the case calls the connection
        # lost method indicating that the transport is now
        # closed (no connection available)
        if not self._protocol == None:
            self._protocol.connection_lost(context)

        # forces the unset of the protocol as the proper
        # connection lost operation has been called and
        # there's no more logical association between
        # the transport and the protocol
        self._protocol = None

    def _call_soon(self, callback, *args):
        if hasattr(self._loop, "call_soon"):
            self._loop.call_soon(callback, *args)
        else:
            callable = lambda: callback(*args)
            self._loop.delay(callable, immediately = True)

class TransportDatagram(Transport):
    """
    Abstract class to be used when creating a datagram based
    (connectionless) transport.

    This implementation reflects the decisions made for the
    netius based transport abstraction, inspired by asyncio.
    """

    def _on_data(self, connection, data):
        data, address = data
        self._protocol.datagram_received(data, address)

    def _on_close(self, connection):
        self._cleanup()

class TransportStream(Transport):
    """
    Abstract class to be used when creating a stream based
    (connection) transport.

    This implementation reflects the decisions made for the
    netius based transport abstraction, inspired by asyncio.
    """

    def _on_data(self, connection, data):
        self._protocol.data_received(data)

    def _on_close(self, connection):
        if not self._protocol == None:
            self._protocol.eof_received()
        self._cleanup()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import re
import os
import ssl
import hashlib

from . import config
from . import errors

def fingerprint(certificate, hash = "sha1"):
    digest = hashlib.new(hash, certificate)
    return digest.hexdigest()

def match_fingerprint(certificate, exp_fingerprint, hash = "sha1"):
    cert_fingerprint = fingerprint(certificate, hash = hash)
    if cert_fingerprint == exp_fingerprint: return
    if config._is_devel():
        extra = ", expected '%s' got '%s'" %\
            (exp_fingerprint, cert_fingerprint)
    else:
        extra = ""
    raise errors.SecurityError(
        "Missmatch in certificate fingerprint" + extra
    )

def match_hostname(certificate, hostname):
    if hasattr(ssl, "match_hostname"):
        return ssl.match_hostname(certificate, hostname)

    dns_names = []
    subject_alt_name = certificate.get("subjectAltName", ())

    for key, value in subject_alt_name:
        if not key == "DNS": continue
        if dnsname_match(value, hostname): return
        dns_names.append(value)

    if not dns_names:
        for subject in certificate.get("subject", ()):
            for key, value in subject:
                if not key == "commonName": continue
                if dnsname_match(value, hostname): return
                dns_names.append(value)

    if len(dns_names) > 1:
        raise errors.SecurityError(
            "Hostname %s doesn't match either of %s" %\
            (hostname, ", ".join(map(str, dns_names)))
        )
    elif len(dns_names) == 1:
        raise errors.SecurityError(
            "Hostname %s doesn't match %s" %\
            (hostname, dns_names[0])
        )
    else:
        raise errors.SecurityError(
            "No appropriate commonName or subjectAltName fields were found"
        )

def dnsname_match(domain, hostname, max_wildcards = 1):
    # creates the initial list of pats that are going to be used in
    # the final match operation for wildcard matching
    pats = []

    # in case no valid domain is passed an invalid result is returned
    # immediately indicating that no match was possible
    if not domain: return False

    # splits the provided domain value around its components, taking
    # into account the typical dot separator
    parts = domain.split(".")
    base = parts[0]
    remainder = parts[1:]

    # determines the number of wildcard characters present in the
    # base value for discovery in case this value overflow the maximum
    # number of wildcards allowed raises an error
    wildcards = base.count("*")
    if wildcards > max_wildcards: raise errors.SecurityError(
        "Too many wildcards in certificate DNS name: " + str(domain)
    )

    # in case there are no wildcards in the domain name runs the
    # "normal" hostname validation process against the domain name
    if wildcards == 0:
        return domain.lower() == hostname.lower()

    if base == "*":
        pats.append("[^.]+")
    elif base.startswith("xn--") or hostname.startswith("xn--"):
        pats.append(re.escape(base))
    else:
        pats.append(re.escape(base).replace(r"\*", "[^.]*"))

    for fragment in remainder:
        pats.append(re.escape(fragment))

    pat = re.compile(r"\A" + r"\.".join(pats) + r"\Z", re.IGNORECASE)
    return True if pat.match(hostname) else False

def dump_certificate(certificate, certificate_binary, name = None):
    # runs some pre-validation operations so that the dump parameters
    # that are required are considered valid
    if not certificate: return
    if not certificate_binary: return

    # tries to retrieve the main subject name from the subject
    # alternative names, there may be no value and if that's the
    # case a default value is used instead
    subject_alt_name = certificate.get("subjectAltName", ())
    if subject_alt_name: subject_name = subject_alt_name[0][1]
    else: subject_name = "certificate"

    # "calculates" the final name for the certificate, taking
    # into account the provided parameter and subject name, the
    # constructs the file name value with the der suffix
    name = name or subject_name
    file_name = name + ".der"

    # retrieves the proper ssl path (where data is going to be
    # stored) and creates such directory (if required)
    ssl_path = config.conf("SSL_PATH", "/tmp/ssl")
    file_path = os.path.join(ssl_path, file_name)
    if not os.path.exists(ssl_path): os.makedirs(ssl_path)

    # opens the file for writing and then dumps the certificate
    # binary information into the file, closing it afterwards
    file = open(file_path, "wb")
    try: file.write(certificate_binary)
    finally: file.close()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import uuid
import hashlib

import netius

class BaseAdapter(object):
    """
    Top level abstract representation of a netius adapter.
    The adapter is responsible for the storage/retrieval
    of key based values, and may be used as the storage
    engine back-end for a series or purposes (eg: email
    storage, torrent hash table storage, sessions, etc.)
    """

    def set(self, value, owner = "nobody"):
        pass

    def get(self, key):
        file = self.get_file(key)
        if not file: return file
        try: value = file.read()
        finally: file.close()
        return value

    def get_file(self, key, mode = "rb"):
        return netius.legacy.StringIO()

    def delete(self, key, owner = "nobody"):
        pass

    def append(self, key, value):
        file = self.get_file(key, mode = "ab")
        try: file.write(value)
        finally: file.close()

    def truncate(self, key, count):
        file = self.get_file(key, mode = "rb+")
        try:
            offset = count * -1
            file.seek(offset, os.SEEK_END)
            file.truncate()
        finally:
            file.close()

    def size(self, key):
        return 0

    def sizes(self, owner = None):
        list = self.list(owner = owner)
        sizes = [self.size(key) for key in list]
        return sizes

    def total(self, owner = None):
        total = 0
        list = self.list(owner = owner)
        for key in list: total += self.size(key)
        return total

    def reserve(self, owner = "nobody"):
        return self.set("", owner = owner)

    def count(self, owner = None):
        return 0

    def list(self, owner = None):
        return ()

    def generate(self):
        identifier = str(uuid.uuid4())
        identifier = netius.legacy.bytes(identifier)
        hash = hashlib.sha256(identifier)
        key = hash.hexdigest()
        return key



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import math
import socket
import collections

import netius

SIZE_UNITS_LIST = (
    "B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"
)
""" The size units list that contains the complete set of
units indexed by the depth they represent """

SIZE_UNITS_LIST_S = (
    "B", "K", "M", "G", "T", "P", "E", "Z", "Y"
)
""" The simplified size units list that contains the complete set of
units indexed by the depth they represent """

SIZE_UNIT_COEFFICIENT = 1024
""" The size unit coefficient as an integer value, this is
going to be used in each of the size steps as divisor """

DEFAULT_MINIMUM = 1024
""" The default minimum value meaning that this is the
maximum value that one integer value may have for the
size rounding operation to be performed """

DEFAULT_PLACES = 3
""" The default number of places (digits) that are going
to be used for the string representation in the round
based conversion of size units to be performed """

_HOST = None
""" The globally cached value for the current hostname,
this value is used to avoid an excessive blocking in the
get host by name call, as it is a blocking call """

def cstring(value):
    index = value.index("\0")
    if index == -1: return value
    return value[:index]

def chunks(sequence, count):
    for index in range(0, len(sequence), count):
        yield sequence[index:index + count]

def header_down(name):
    values = name.split("-")
    values = [value.lower() for value in values]
    return "-".join(values)

def header_up(name):
    values = name.split("-")
    values = [value.title() for value in values]
    return "-".join(values)

def is_ip4(address):
    address_p = address.split(".", 4)
    if not len(address_p) == 4: return False
    for part in address_p:
        try: part_i = int(part)
        except ValueError: return False
        if part_i < 0: return False
        if part_i > 255: return False
    return True

def is_ip6(address):
    if is_ip4(address): return False
    return True

def assert_ip4(address, allowed, default = True):
    if not allowed: return default
    for item in allowed:
        is_subnet = "/" in item
        if is_subnet: valid = in_subnet_ip4(address, item)
        else: valid = address == item
        if not valid: continue
        return True
    return False

def in_subnet_ip4(address, subnet):
    subnet, length = subnet.split("/", 1)
    size_i = 32 - int(length)
    address_a = ip4_to_addr(address)
    subnet_a = ip4_to_addr(subnet)
    limit_a = subnet_a + pow(2, size_i)
    in_subnet = (address_a & subnet_a) == subnet_a
    in_subnet &= address_a < limit_a
    return in_subnet

def addr_to_ip4(number):
    first = int(number / 16777216) % 256
    second = int(number / 65536) % 256
    third = int(number / 256) % 256
    fourth = int(number) % 256
    return "%s.%s.%s.%s" % (first, second, third, fourth)

def addr_to_ip6(number):
    buffer = collections.deque()
    for index in range(8):
        offset = index * 2
        first = number >> (8 * offset) & 0xff
        second = number >> (8 * (offset + 1)) & 0xff
        buffer.appendleft("%02x%02x" % (second, first))
    return ":".join(buffer)

def ip4_to_addr(value):
    first, second, third, fourth = value.split(".", 3)
    first_a = int(first) * 16777216
    second_a = int(second) * 65536
    third_a = int(third) * 256
    fourth_a = int(fourth)
    return first_a + second_a + third_a + fourth_a

def string_to_bits(value):
    return bin(netius.legacy.reduce(lambda x, y : (x << 8) + y, (netius.legacy.ord(c) for c in value), 1))[3:]

def integer_to_bytes(number, length = 0):
    if not isinstance(number, netius.legacy.INTEGERS):
        raise netius.DataError("Invalid data type")

    bytes = []
    number = abs(number)

    while number > 0:
        bytes.append(chr(number & 0xff))
        number >>= 8

    remaining = length - len(bytes)
    remaining = 0 if remaining < 0 else remaining
    for _index in range(remaining): bytes.append("\x00")

    bytes = reversed(bytes)
    bytes_s = "".join(bytes)
    bytes_s = netius.legacy.bytes(bytes_s)

    return bytes_s

def bytes_to_integer(bytes):
    if not type(bytes) == netius.legacy.BYTES:
        raise netius.DataError("Invalid data type")

    number = 0
    for byte in bytes: number = (number << 8) | netius.legacy.ord(byte)
    return number

def random_integer(number_bits):
    """
    Generates a random integer of approximately the
    size of the provided number bits bits rounded up
    to whole bytes.

    :type number_bits: int
    :param number_bits: The number of bits of the generated
    random integer, this value will be used as the basis
    for the calculus of the required bytes.
    :rtype: int
    :return: The generated random integer, should be provided
    with the requested size.
    """

    # calculates the number of bytes to represent the number
    # by dividing the number of bits by a byte and then rounding
    # the value to the next integer value
    number_bytes = math.ceil(number_bits / 8.0)
    number_bytes = int(number_bytes)

    # generates a random data string with the specified
    # number of bytes in length
    random_data = os.urandom(number_bytes)

    # converts the random data into an integer and then
    # makes sure the last bit of the value is correctly
    # filled with data, and returns it to the caller method
    random_integer = bytes_to_integer(random_data)
    random_integer |= 1 << (number_bits - 1)
    return random_integer

def host(default = "127.0.0.1"):
    """
    Retrieves the host for the current machine,
    typically this would be the ipv4 address of
    the main network interface.

    No result type are guaranteed and a local address
    (eg: 127.0.0.1) may be returned instead.

    The returned value is cached to avoid multiple
    blocking calls from blocking the processor.

    :type default: String
    :param default: The default value that is going to
    be returned in case no resolution is possible, take
    into account that this result is going to be cached.
    :rtype: Strong
    :return: The string that contains the host address
    as defined by specification for the current machine.
    """

    global _HOST
    if _HOST: return _HOST
    hostname = socket.gethostname()
    try: _HOST = socket.gethostbyname(hostname)
    except socket.gaierror: _HOST = default
    is_unicode = type(_HOST) == netius.legacy.OLD_UNICODE
    if is_unicode: _HOST = _HOST.encode("utf-8")
    return _HOST

def hostname():
    """
    The name as a simple string o the name of the current
    local machine. This value may or may not be a fully
    qualified domain name for the machine.

    The result of this function call is unpredictable and
    should not be trusted for critical operations.

    :rtype: String
    :return: The name as a string of the current local
    machine, the definition of this value varies.
    """

    return socket.gethostname()

def size_round_unit(
    size_value,
    minimum = DEFAULT_MINIMUM,
    places = DEFAULT_PLACES,
    reduce = True,
    space = False,
    justify = False,
    simplified = False,
    depth = 0
):
    """
    Rounds the size unit, returning a string representation
    of the value with a good rounding precision.
    This method should be used to round data sizing units.

    Note that using the places parameter it's possible to control
    the number of digits (including decimal places) of the
    number that is going to be "generated".

    :type size_value: int/float
    :param size_value: The current size value (in bytes).
    :type minimum: int
    :param minimum: The minimum value to be used.
    :type places: int
    :param places: The target number of digits to be used for
    describing the value to be used for output, this is going
    to be used to calculate the proper number of decimal places.
    :type reduce: bool
    :param reduce: If the final string value should be reduced
    meaning that right decimal zeros should be removed as they
    represent an extra unused value.
    :type space: bool
    :param space: If a space character must be used dividing
    the value from the unit symbol.
    :type justify: bool
    :param justify: If the size string value should be (right)
    justified important for properly aligned values in a table.
    :type simplified: bool
    :param simplified: If the simplified version of the units
    should be used instead of the longer one.
    :type depth: int
    :param depth: The current iteration depth value.
    :rtype: String
    :return: The string representation of the data size
    value in a simplified manner (unit).
    """

    # in case the current size value is acceptable (less than
    # the minimum) this is the final iteration and the final
    # string representation is going to be created
    if size_value < minimum:
        # calculates the maximum size of the string that is going
        # to represent the base size value as the number of places
        # plus one (representing the decimal separator character)
        size_s = places + 1

        # calculates the target number of decimal places taking
        # into account the size (in digits) of the current size
        # value, this may never be a negative number
        log_value = size_value and math.log10(size_value)
        digits = int(log_value) + 1
        places = places - digits
        places = places if places > 0 else 0

        # creates the proper format string that is going to
        # be used in the creation of the proper float value
        # according to the calculated number of places
        format = "%%.%df" % places

        # rounds the size value, then converts the rounded
        # size value into a string based representation
        size_value = round(size_value, places)
        size_value_s = format % size_value

        # forces the reduce flag when the depth is zero, meaning
        # that an integer value will never be decimal, this is
        # required to avoid strange results for depth zero
        reduce = reduce or depth == 0

        # in case the dot value is not present in the size value
        # string adds it to the end otherwise an issue may occur
        # while removing extra padding characters for reduce
        if reduce and not "." in size_value_s: size_value_s += "."

        # strips the value from zero appended to the right and
        # then strips the value also from a possible decimal
        # point value that may be included in it, this is only
        # performed in case the reduce flag is enabled
        if reduce: size_value_s = size_value_s.rstrip("0")
        if reduce: size_value_s = size_value_s.rstrip(".")

        # in case the justify flag is set runs the justification
        # process on the size value taking into account the maximum
        # size of the associated size string
        if justify: size_value_s = size_value_s.rjust(size_s)

        # retrieves the size unit (string mode) for the current
        # depth according to the provided map
        if simplified: size_unit = SIZE_UNITS_LIST_S[depth]
        else: size_unit = SIZE_UNITS_LIST[depth]

        # retrieves the appropriate separator based
        # on the value of the space flag
        separator = space and " " or ""

        # creates the size value string appending the rounded
        # size value string and the size unit and returns it
        # to the caller method as the size value string
        size_value_string = size_value_s + separator + size_unit
        return size_value_string

    # otherwise the value is not acceptable and a new iteration
    # must be ran with one less depth of size value
    else:
        # re-calculates the new size value, increments the depth
        # and runs the size round unit again with the new values
        new_size_value = float(size_value) / SIZE_UNIT_COEFFICIENT
        new_depth = depth + 1
        return size_round_unit(
            new_size_value,
            minimum = minimum,
            places = places,
            reduce = reduce,
            space = space,
            justify = justify,
            simplified = simplified,
            depth = new_depth
        )

def verify(condition, message = None, exception = None):
    """
    Ensures that the requested condition returns a valid value
    and if that's no the case an exception raised breaking the
    current execution logic.

    :type condition: bool
    :param condition: The condition to be evaluated and that may
    trigger an exception raising.
    :type message: String
    :param message: The message to be used in the building of the
    exception that is going to be raised in case of condition failure.
    :type exception: Class
    :param exception: The exception class that is going to be used
    to build the exception to be raised in case the condition
    verification operation fails.
    """

    if condition: return
    exception = exception or netius.AssertionError
    raise exception(message or "Assertion Error")

def verify_equal(first, second, message = None, exception = None):
    message = message or "Expected %s got %s" % (repr(second), repr(first))
    return verify(
        first == second,
        message = message,
        exception = exception
    )

def verify_not_equal(first, second, message = None, exception = None):
    message = message or "Expected %s not equal to %s" % (repr(first), repr(second))
    return verify(
        not first == second,
        message = message,
        exception = exception
    )

def verify_many(sequence, message = None, exception = None):
    for condition in sequence:
        verify(
            condition,
            message = message,
            exception = exception
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import re
import struct
import hashlib

import netius

from . import parser

HANDSHAKE_SIZE = 68
""" The typical size for an handshake message according
to the current torrent protocol definition, the value
is defined in bytes """

DECIMAL_REGEX = re.compile(r"\d")
""" Simple regular expression that matched any
character and expression with decimal digits """

TORRENT_TYPES = {
    -1 : "keep-alive",
    0 : "choke",
    1 : "unchoke",
    2 : "interested",
    3 : "not interested",
    4 : "have",
    5 : "bitfield",
    6 : "request",
    7 : "piece",
    8 : "cancel",
    9 : "port"
}
""" The map that associates the various message type
identifiers with their internal string representations """

def info_hash(root):
    info = root["info"]
    data = bencode(info)
    info_hash = hashlib.sha1(data)
    return info_hash.digest()

def bencode(root):
    # joins the complete set of values created by
    # generator that has been returned from the chunk
    # function using the provided root value
    data = b"".join([value for value in chunk(root)])
    return data

def bdecode(data):
    # converts the provide (string) data into a list
    # of chunks (characters) reversing it so that the
    # proper pop/push operations may be performed, as
    # pop is done from the back of the list
    data = netius.legacy.str(data)
    chunks = list(data)
    chunks.reverse()

    # runs the dechunking operation in the created list
    # of chunks obtaining a dictionary base structure
    # as the result of the bencoding operation
    root = dechunk(chunks)
    return root

def chunk(item):
    chunk_t = type(item)

    if chunk_t == bytes:
        item = netius.legacy.str(item)
        chunk_t = type(item)

    if chunk_t == dict:
        yield b"d"
        keys = item.keys()
        keys = list(keys)
        keys.sort()
        for key in keys:
            value = item[key]
            for part in chunk(key): yield part
            for part in chunk(value): yield part
        yield b"e"

    elif chunk_t == list:
        yield b"l"
        for value in item:
            for part in chunk(value): yield part
        yield b"e"

    elif chunk_t in netius.legacy.INTEGERS:
        yield netius.legacy.bytes("i%de" % item)

    elif chunk_t in netius.legacy.STRINGS:
        yield netius.legacy.bytes("%d:%s" % (len(item), item))

    else:
        raise netius.ParserError("Not possible to encode")

def dechunk(chunks):
    item = chunks.pop()

    if item == "d":
        item = chunks.pop()
        hash = {}

        while not item == "e":
            chunks.append(item)
            key = dechunk(chunks)
            hash[key] = dechunk(chunks)
            item = chunks.pop()

        return hash

    elif item == "l":
        item = chunks.pop()
        list = []

        while not item == "e":
            chunks.append(item)
            list.append(dechunk(chunks))
            item = chunks.pop()

        return list

    elif item == "i":
        item = chunks.pop()
        number = ""

        while not item == "e":
            number += item
            item = chunks.pop()

        return int(number)

    elif DECIMAL_REGEX.search(item):
        number = ""

        while DECIMAL_REGEX.search(item):
            number += item
            item = chunks.pop()

        line = []
        number = int(number)

        for _index in range(number):
            char = chunks.pop()
            line.append(char)

        line = "".join(line)

        return line

    raise netius.ParserError("Invalid input: '%s'" % item)

class TorrentParser(parser.Parser):

    def __init__(self, owner, store = False):
        parser.Parser.__init__(self, owner)

        self.length = None
        self.buffer = []
        self.buffer_l = 0

        self.build()

    def build(self):
        """
        Builds the initial set of states ordered according to
        their internal integer definitions, this method provides
        a fast and scalable way of parsing data.
        """

        parser.Parser.build(self)

        self.states = (
            self._parse_handshake,
            self._parse_message
        )
        self.state_l = len(self.states)

    def destroy(self):
        """
        Destroys the current structure for the parser meaning that
        it's restored to the original values, this method should only
        be called on situation where no more parser usage is required.
        """

        parser.Parser.destroy(self)

        self.states = ()
        self.state_l = 0

    def parse(self, data):
        """
        Parses the provided data chunk, changing the current
        state of the parser accordingly and returning the
        number of processed bytes from it.

        :type data: String
        :param data: The string containing the data to be parsed
        in the current parse operation.
        :rtype: int
        :return: The amount of bytes of the data string that have
        been "parsed" in the current parse operation.
        """

        parser.Parser.parse(self, data)

        # retrieves the size of the data that has been sent for parsing
        # and saves it under the size original variable
        size = len(data)
        size_o = size

        # iterates continuously to try to process all that
        # data that has been sent for processing
        while size > 0:
            # in case there's no owner associated with the
            # current parser must break the loop because
            # there's no way to continue with parsing
            if not self.owner: break

            # retrieves the parsing method for the current
            # state and then runs it retrieving the number
            # of valid parsed bytes in case this value is
            # zero the parsing iteration is broken
            method = self.states[self.owner.state - 1]
            count = method(data)
            if count == 0: break

            # decrements the size of the data buffer by the
            # size of the parsed bytes and then retrieves the
            # sub part of the data buffer as the new data buffer
            size -= count
            data = data[count:]

        # in case not all of the data has been processed
        # must add it to the buffer so that it may be used
        # latter in the next parsing of the message
        if size > 0: self.buffer.append(data); self.buffer_l += size

        # returns the number of read (processed) bytes of the
        # data that has been sent to the parser
        return size_o - size

    def _join(self, data):
        self.buffer.append(data)
        result = b"".join(self.buffer)
        self.buffer = [result]
        self.buffer_l = len(result)
        return result

    def _parse_handshake(self, data):
        total = len(data) + self.buffer_l

        if total < HANDSHAKE_SIZE: return 0

        diff = HANDSHAKE_SIZE - self.buffer_l if self.buffer_l < HANDSHAKE_SIZE else 0
        result = self._join(data[:diff])

        _length, protocol, reserved, info_hash, peer_id = struct.unpack("!B19sQ20s20s", result)
        self.trigger("on_handshake", protocol, reserved, info_hash, peer_id)

        self.length = None
        del self.buffer[:]
        self.buffer_l = 0

        return diff

    def _parse_message(self, data):
        # starts the current processed byte count at zero and then
        # measures the current total received bytes using the length
        # of the current data chunk received and the buffer length
        count = 0
        total = len(data) + self.buffer_l

        # in case the length is not yet defined it must be searched
        # inside the current message so that it may than be re-used
        # to check if the complete message has been received
        if self.length == None:
            if total < 4: return 0
            diff = 4 - self.buffer_l if self.buffer_l < 4 else 0
            result = self._join(data[:diff])
            data = data[diff:]
            self.length, = struct.unpack("!L", result[:4])
            count += diff

        # calculates the "target" total message length and verifies
        # that the total number of bytes received for the message
        # have already reached that value otherwise returns count
        # to the caller method (delayed execution)
        message_length = self.length + 4
        if total < message_length: return count

        # calculates the difference between meaning the amount of
        # data from the current chunk that is going to be processed
        # and then joins the current buffer accordingly
        diff = message_length - self.buffer_l if self.buffer_l < message_length else 0
        result = self._join(data[:diff])

        # verifies if the message contains a payload (length is
        # greater than zero) for such situations the type must
        # be loaded, otherwise the type is assumed to be the keep
        # alive one (only message with no payload available)
        if self.length > 0: type, = struct.unpack("!B", result[4:5])
        else: type = -1

        # resolves the current type integer based type into the proper
        # string based type values so that it may be used from now on
        type_s = TORRENT_TYPES.get(type, "invalid")
        data = result[5:]

        # triggers the on message event notifying the listeners about
        # the current message that has been received (includes payload and type)
        self.trigger("on_message", self.length, type_s, data)

        # resets the current message processing state as the message
        # has been completely processed, new message may come and require
        # a newly created state for the parser (to be correctly processed)
        self.length = None
        del self.buffer[:]
        self.buffer_l = 0

        # increments the current byte processed counter with the current
        # diff value and then returns the count value to the caller method
        # notifying it about the number of processed bytes
        count += diff
        return count



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json
import struct
import binascii

import netius

class APNProtocol(netius.StreamProtocol):
    """
    Protocol class that defines the interface to operate
    an apple push notifications (APN) protocol that is
    able to send push notifications to Apple devices.

    Should be compliant with a simple message oriented
    interface for easy of usage from developers.
    """

    HOST = "gateway.push.apple.com"
    """ The host of the APN service to be used when
    in production mode """

    PORT = 2195
    """ The port of the APN service to be used when
    in sandbox mode """

    SANDBOX_HOST = "gateway.sandbox.push.apple.com"
    """ The host of the APN service to be used when
    in sandbox mode (for testing purposes) """

    SANDBOX_PORT = 2195
    """ The port of the APN service to be used when
    in sandbox mode (for testing purposes) """

    def __init__(self, *args, **kwargs):
        netius.StreamProtocol.__init__(self, *args, **kwargs)
        self.host = None
        self.port = None
        self.token = None
        self.message = None
        self.sound = None
        self.badge = 0
        self.sandbox = True
        self._close = True

    def connection_made(self, transport):
        netius.StreamProtocol.connection_made(self, transport)

        self.send_notification(
            self.token,
            self.message,
            sound = self.sound,
            badge = self.badge,
            close = self._close
        )

    def send_notification(
        self,
        token,
        message,
        sound = "default",
        badge = 0,
        close = False
    ):
        # creates the callback handler that closes the current
        # client infra-structure after sending, this will close
        # the connection using a graceful approach to avoid any
        # of the typical problems with the connection shutdown
        def callback(transport): self.close()

        # converts the current token (in hexadecimal) to a set
        # of binary string elements and uses that value to get
        # a string of data from the string of hexadecimal data
        token = netius.legacy.bytes(token)
        token = binascii.unhexlify(token)

        # creates the message structure using with the
        # message (string) as the alert and then converts
        # it into a json format (payload)
        message_s = dict(
           aps = dict(
                alert = message,
                sound = sound,
                badge = badge
            )
        )
        payload = json.dumps(message_s)

        # verifies if the resulting payload object is unicode based
        # and in case it is encodes it into a string representation
        # so that it may be used for the packing of structure
        is_unicode = netius.legacy.is_unicode(payload)
        if is_unicode: payload = payload.encode("utf-8")

        # sets the command with the zero value (simplified)
        # then calculates the token and payload lengths
        command = 0
        token_length = len(token)
        payload_length = len(payload)

        # creates the initial template for message creation by
        # using the token and the payload length for it, then
        # applies the various components of the message and packs
        # them according to the generated template
        template = "!BH%dsH%ds" % (token_length, payload_length)
        message = struct.pack(template, command, token_length, token, payload_length, payload)
        callback = callback if close else None
        self.send(message, callback = callback)

    def set(
        self,
        token,
        message,
        sound = "default",
        badge = 0,
        sandbox = True,
        key_file = None,
        cer_file = None,
        _close = True
    ):
        self.token = token
        self.message = message
        self.sound = sound
        self.badge = badge
        self.sandbox = sandbox
        self.key_file = key_file
        self.cer_file = cer_file
        self._close = _close

    def notify(self, token, loop = None, **kwargs):
        # retrieves the intance's parent class object to be
        # used to global class operations
        cls = self.__class__

        # unpacks the various keyword based arguments for the
        # provided map of arguments defaulting to a series of
        # pre-defined values in case the arguments have not
        # been correctly provided
        message = kwargs.get("message", "Hello World")
        sound = kwargs.get("sound", "default")
        badge = kwargs.get("badge", 0)
        sandbox = kwargs.get("sandbox", True)
        key_file = kwargs.get("key_file", None)
        cer_file = kwargs.get("cer_file", None)
        _close = kwargs.get("close", True)

        # retrieves the values that are going to be used for
        # both the host and the port, taking into account if
        # the current message is meant to be send using the
        # sandbox environment (for testing purposes)
        self.host = cls.SANDBOX_HOST if sandbox else cls.HOST
        self.port = cls.SANDBOX_PORT if sandbox else cls.PORT

        # sets the APN specific information in the current
        # protocol instance to be used once connected
        self.set(
            token,
            message,
            sound = sound,
            badge = badge,
            sandbox = sandbox,
            key_file = key_file,
            cer_file = cer_file,
            _close = _close
        )

        # establishes the connection to the target host and port
        # and using the provided key and certificate files
        loop = netius.connect_stream(
            lambda: self,
            host = self.host,
            port = self.port,
            ssl = True,
            key_file = key_file,
            cer_file = cer_file,
            loop = loop
        )

        # returns both the current associated loop and the current
        # instance to the protocol defined by the current instance
        return loop, self

class APNClient(netius.ClientAgent):

    protocol = APNProtocol

    @classmethod
    def notify_s(cls, token, loop = None, **kwargs):
        protocol = cls.protocol()
        return protocol.notify(token, loop = loop, **kwargs)

if __name__ == "__main__":
    def on_finish(protocol):
        netius.compat_loop(loop).stop()

    token = netius.conf("APN_TOKEN", None)
    key_file = netius.conf("APN_KEY_FILE", None)
    cer_file = netius.conf("APN_CER_FILE", None)

    loop, protocol = APNClient.notify_s(
        token, key_file = key_file, cer_file = cer_file
    )

    protocol.bind("finish", on_finish)

    loop.run_forever()
    loop.close()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import agent
from . import asynchronous
from . import client
from . import common
from . import compat
from . import config
from . import conn
from . import container
from . import errors
from . import legacy
from . import log
from . import observer
from . import poll
from . import protocol
from . import request
from . import server
from . import stream
from . import tls
from . import transport
from . import util

from .agent import Agent, ClientAgent, ServerAgent
from .asynchronous import Future, Task, Handle, Executor, ThreadPoolExecutor, coroutine,\
    async_test_all, async_test, ensure_generator, get_asyncio, is_coroutine,\
    is_coroutine_object, is_coroutine_native, is_future, is_neo, is_asynclib, is_await,\
    wakeup, sleep, wait, notify, coroutine_return
from .client import Client, DatagramClient, StreamClient
from .common import NAME, VERSION, IDENTIFIER_SHORT, IDENTIFIER_LONG,\
    IDENTIFIER, TCP_TYPE, UDP_TYPE, SSL_KEY_PATH, SSL_CER_PATH, SSL_CA_PATH,\
    SSL_DH_PATH, Base, BaseThread, new_loop_main, new_loop_asyncio, new_loop,\
    ensure_main, ensure_asyncio, ensure_loop, get_main, get_loop, get_event_loop,\
    stop_loop, compat_loop, get_poll, build_future, ensure, ensure_pool
from .compat import BaseLoop, CompatLoop, is_compat, is_asyncio, build_datagram,\
    connect_stream
from .config import conf, conf_prefix, conf_suffix, conf_s, conf_r, conf_d, conf_ctx
from .conn import OPEN, CLOSED, PENDING, CHUNK_SIZE, Connection
from .container import Container, ContainerServer
from .errors import NetiusError, RuntimeError, StopError, PauseError, WakeupError,\
    DataError, ParserError, GeneratorError, SecurityError, NotImplemented, AssertionError
from .log import SILENT, rotating_handler, smtp_handler
from .observer import Observable
from .poll import Poll, EpollPoll, KqueuePoll, PollPoll, SelectPoll
from .protocol import Protocol, DatagramProtocol, StreamProtocol
from .request import Request, Response
from .server import Server, DatagramServer, StreamServer
from .stream import Stream
from .tls import fingerprint, match_fingerprint, match_hostname, dnsname_match,\
    dump_certificate
from .transport import Transport, TransportDatagram, TransportStream
from .util import camel_to_underscore, verify



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius

from . import common

FILE_WORK = 20

ERROR_ACTION = -1
OPEN_ACTION = 1
CLOSE_ACTION = 2
READ_ACTION = 3
WRITE_ACTION = 4

class FileThread(common.Thread):

    def execute(self, work):
        type = work[0]
        if not type == FILE_WORK: netius.NotImplemented(
            "Cannot execute type '%d'" % type
        )

        try:
            self._execute(work)
        except BaseException as exception:
            self.owner.push_event((ERROR_ACTION, exception, work[-1]))

    def open(self, path, mode, data):
        file = open(path)
        self.owner.push_event((OPEN_ACTION, file, data))

    def close(self, file, data):
        file.close()
        self.owner.push_event((CLOSE_ACTION, file, data))

    def read(self, file, count, data):
        result = file.read(count)
        self.owner.push_event((READ_ACTION, result, data))

    def write(self, file, buffer, data):
        file.write(buffer)
        self.owner.push_event((WRITE_ACTION, len(buffer), data))

    def _execute(self, work):
        action = work[1]
        if action == OPEN_ACTION: self.open(*work[2:])
        elif action == CLOSE_ACTION: self.close(*work[2:])
        elif action == READ_ACTION: self.read(*work[2:])
        elif action == WRITE_ACTION: self.read(*work[2:])
        else: netius.NotImplemented("Undefined file action '%d'" % action)

class FilePool(common.EventPool):

    def __init__(self, base = FileThread, count = 10):
        common.EventPool.__init__(self, base = base, count = count)

    def open(self, path, mode = "r", data = None):
        work = (FILE_WORK, OPEN_ACTION, path, mode, data)
        self.push(work)

    def close(self, file, data = None):
        work = (FILE_WORK, CLOSE_ACTION, file, data)
        self.push(work)

    def read(self, file, count = -1, data = None):
        work = (FILE_WORK, READ_ACTION, file, count, data)
        self.push(work)

    def write(self, file, buffer, data = None):
        work = (FILE_WORK, WRITE_ACTION, file, buffer, data)
        self.push(work)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius

from . import util

INTEGER = 0x02
BIT_STRING = 0x03
OCTET_STRING = 0x04
NULL = 0x05
OBJECT_IDENTIFIER = 0x06
SEQUENCE = 0x30

ASN1_OBJECT = [
    (SEQUENCE, [
        (SEQUENCE, [
            OBJECT_IDENTIFIER,
            NULL
        ]),
        BIT_STRING
    ])
]

ASN1_RSA_PUBLIC_KEY = [
    (SEQUENCE, [
        INTEGER,
        INTEGER
    ])
]

ASN1_RSA_PRIVATE_KEY = [
    (SEQUENCE, [
        INTEGER,
        INTEGER,
        INTEGER,
        INTEGER,
        INTEGER,
        INTEGER,
        INTEGER,
        INTEGER,
        INTEGER
    ])
]

RSAID_PKCS1 = b"\x2a\x86\x48\x86\xf7\x0d\x01\x01\x01"
HASHID_SHA1 = b"\x2b\x0e\x03\x02\x1a"
HASHID_SHA256 = b"\x60\x86\x48\x01\x65\x03\x04\x02\x01"

def asn1_parse(template, data):
    """
    Parse a data structure according to asn.1 template,
    the provided template should respect the predefined
    structure. The provided data is going to be validated
    against the template format and a exception raised in
    case the data is not conformant.

    :type template: List/Tuple
    :param template: A list of tuples comprising the asn.1 template.
    :type: List
    :param data: A list of bytes to parse.
    """

    # starts the index value at the zero value and the creates the
    # the list that is going to hold the various partial values for
    # generated during the parsing of the structure
    index = 0
    result = []

    # iterates over the complete set of items in the current
    # template structure to "apply" them to the current values
    # so that they are correctly parsed/verified
    for item in template:
        # verifies if the data type for the current template
        # item to be parser is tuple and based on that defined
        # the current expected data type and children values
        is_tuple = type(item) == tuple
        if is_tuple: dtype, children = item
        else: dtype = item; children = None

        # retrieves the value (as an ordinal) for the current
        # byte and increments the index for the parser
        tag = netius.legacy.ord(data[index])
        index += 1

        # in case the current type is not of the expect type,
        # must raise an exception indicating the problem to
        # the top level layers (should be properly handled)
        if not tag == dtype:
            raise netius.ParserError("Unexpected tag (got 0x%02x, expecting 0x%02x)" % (tag, dtype))

        # retrieves the ordinal value of the current byte as
        # the length of the value to be parsed and then increments
        # the pointer of the buffer reading process
        length = netius.legacy.ord(data[index])
        index += 1

        # in case the last bit of the length byte is set the,
        # the byte designates the length of the byte sequence that
        # defines the length of the current value to be read instead
        if length & 0x80:
            number = length & 0x7f
            length = util.bytes_to_integer(data[index:index + number])
            index += number

        if tag == BIT_STRING:
            result.append(data[index:index + length])
            index += length

        elif tag == OCTET_STRING:
            result.append(data[index:index + length])
            index += length

        elif tag == INTEGER:
            number = util.bytes_to_integer(data[index:index + length])
            index += length
            result.append(number)

        elif tag == NULL:
            netius.verify(length == 0)
            result.append(None)

        elif tag == OBJECT_IDENTIFIER:
            result.append(data[index:index + length])
            index += length

        elif tag == SEQUENCE:
            part = asn1_parse(children, data[index:index + length])
            result.append(part)
            index += length

        else:
            raise netius.ParserError("Unexpected tag in template 0x%02x" % tag)

    return result

def asn1_length(length):
    """
    Returns a string representing a field length in asn.1 format.
    This value is computed taking into account the multiple byte
    representation of very large values.

    :type length: int
    :param length:The integer based length value that is going to
    be used in the conversion to a string representation.
    :rtype: String
    :return: The string based representation of the provided length
    integer value according to the asn.1 specification.
    """

    netius.verify(length >= 0)
    if length < 0x7f: return netius.legacy.chr(length)

    result = util.integer_to_bytes(length)
    number = len(result)
    result = netius.legacy.chr(number | 0x80) + result
    return result

def asn1_gen(node):
    generator = asn1_build(node)
    return b"".join(generator)

def asn1_build(node):
    """
    Builds an asn.1 data structure based on pairs of (type, data),
    this function may be used as a generator of a buffer.

    :type node: Tuple
    :param node: The root node of the structure that is going to be
    used as reference for the generation of the asn.1 buffer.
    """

    tag, value = node

    if tag == BIT_STRING:
        yield netius.legacy.chr(BIT_STRING) + asn1_length(len(value)) + value

    elif tag == OCTET_STRING:
        yield netius.legacy.chr(OCTET_STRING) + asn1_length(len(value)) + value

    elif tag == INTEGER:
        value = util.integer_to_bytes(value)
        yield netius.legacy.chr(INTEGER) + asn1_length(len(value)) + value

    elif tag == NULL:
        netius.verify(value == None)
        yield netius.legacy.chr(NULL) + asn1_length(0)

    elif tag == OBJECT_IDENTIFIER:
        yield netius.legacy.chr(OBJECT_IDENTIFIER) + asn1_length(len(value)) + value

    elif tag == SEQUENCE:
        buffer = []
        for item in value:
            generator = asn1_build(item)
            data = b"".join(generator)
            buffer.append(data)
        result = b"".join(buffer)
        yield netius.legacy.chr(SEQUENCE) + asn1_length(len(result)) + result

    else:
        raise netius.GeneratorError("Unexpected tag in template 0x%02x" % tag)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import math
import base64

import netius

from . import asn
from . import util
from . import calc

PRIVATE_TOKEN = "RSA PRIVATE KEY"
PUBLIC_TOKEN = "PUBLIC KEY"

def open_pem_key(path, token = PRIVATE_TOKEN):
    is_file = not isinstance(path, netius.legacy.STRINGS)
    if is_file: file = path
    else: file = open(path, "rb")
    try:
        data = file.read()
    finally:
        if not is_file: file.close()
    return open_pem_data(data, token = token)

def open_pem_data(data, token = PRIVATE_TOKEN):
    begin, end = pem_limiters(token)

    begin_index = data.find(begin)
    end_index = data.find(end)

    if begin_index == -1: raise netius.ParserError("Invalid key format")
    if end_index == -1: raise netius.ParserError("Invalid key format")

    begin_index += len(begin)

    data = data[begin_index:end_index]
    data = data.strip()
    return base64.b64decode(data)

def write_pem_key(
    path,
    data,
    token = PRIVATE_TOKEN,
    width = 64
):
    begin, end = pem_limiters(token)

    data = base64.b64encode(data)

    chunks = [chunk for chunk in util.chunks(data, width)]
    data = b"\n".join(chunks)

    is_file = not isinstance(path, netius.legacy.STRINGS)
    file = path if is_file else open(path, "wb")
    try:
        file.write(begin)
        file.write(b"\n")
        file.write(data)
        file.write(b"\n")
        file.write(end)
        file.write(b"\n")
    finally:
        if not is_file: file.close()

def open_private_key(path):
    data = open_pem_key(
        path,
        token = PRIVATE_TOKEN
    )
    return open_private_key_data(data)

def open_private_key_b64(data_b64):
    data = base64.b64decode(data_b64)
    return open_private_key_data(data)

def open_private_key_data(data):
    asn1 = asn.asn1_parse(asn.ASN1_RSA_PRIVATE_KEY, data)[0]
    private_key = dict(
        version = asn1[0],
        modulus = asn1[1],
        public_exponent = asn1[2],
        private_exponent = asn1[3],
        prime_1 = asn1[4],
        prime_2 = asn1[5],
        exponent_1 = asn1[6],
        exponent_2 = asn1[7],
        coefficient = asn1[8],
        bits = rsa_bits(asn1[1])
    )
    return private_key

def open_public_key(path):
    data = open_pem_key(
        path,
        token = PUBLIC_TOKEN
    )
    return open_public_key_data(data)

def open_public_key_b64(data_b64):
    data = base64.b64decode(data_b64)
    return open_public_key_data(data)

def open_public_key_data(data):
    asn1 = asn.asn1_parse(asn.ASN1_OBJECT, data)[0]
    asn1 = asn.asn1_parse(asn.ASN1_RSA_PUBLIC_KEY, asn1[1][1:])[0]
    public_key = dict(
        modulus = asn1[0],
        public_exponent = asn1[1],
        bits = rsa_bits(asn1[0])
    )
    return public_key

def write_private_key(path, private_key):
    data = asn_private_key(private_key)
    write_pem_key(
        path,
        data,
        token = PRIVATE_TOKEN
    )

def write_public_key(path, public_key):
    data = asn_public_key(public_key)
    write_pem_key(
        path,
        data,
        token = PUBLIC_TOKEN
    )

def asn_private_key(private_key):
    return asn.asn1_gen(
        (asn.SEQUENCE, [
            (asn.INTEGER, private_key["version"]),
            (asn.INTEGER, private_key["modulus"]),
            (asn.INTEGER, private_key["public_exponent"]),
            (asn.INTEGER, private_key["private_exponent"]),
            (asn.INTEGER, private_key["prime_1"]),
            (asn.INTEGER, private_key["prime_2"]),
            (asn.INTEGER, private_key["exponent_1"]),
            (asn.INTEGER, private_key["exponent_2"]),
            (asn.INTEGER, private_key["coefficient"])
        ])
    )

def asn_public_key(public_key):
    data = b"\x00" + asn.asn1_gen(
        (asn.SEQUENCE, [
            (asn.INTEGER, public_key["modulus"]),
            (asn.INTEGER, public_key["public_exponent"])
        ])
    )
    return asn.asn1_gen(
        (asn.SEQUENCE, [
            (asn.SEQUENCE, [
                (asn.OBJECT_IDENTIFIER, asn.RSAID_PKCS1),
                (asn.NULL, None)
            ]),
            (asn.BIT_STRING, data)
        ])
    )

def pem_to_der(in_path, out_path, token = PRIVATE_TOKEN):
    data = open_pem_key(in_path, token = token)
    file = open(out_path, "wb")
    try: file.write(data)
    finally: file.close()

def pem_limiters(token):
    begin = netius.legacy.bytes("-----BEGIN " + token + "-----")
    end = netius.legacy.bytes("-----END " + token + "-----")
    return (begin, end)

def private_to_public(private_key):
    public_key = dict(
        modulus = private_key["modulus"],
        public_exponent = private_key["public_exponent"]
    )
    return public_key

def assert_private(private_key, number_bits = None):
    prime_1 = private_key["prime_1"]
    prime_2 = private_key["prime_2"]
    private_exponent = private_key["private_exponent"]

    modulus = prime_1 * prime_2
    exponent_1 = private_exponent % (prime_1 - 1)
    exponent_2 = private_exponent % (prime_2 - 1)
    coefficient = calc.modinv(prime_2, prime_1)

    netius.verify(modulus == private_key["modulus"])
    netius.verify(exponent_1 == private_key["exponent_1"])
    netius.verify(exponent_2 == private_key["exponent_2"])
    netius.verify(coefficient == private_key["coefficient"])

    if number_bits:
        netius.verify(number_bits // 2 == rsa_bits(private_key["prime_1"]))
        netius.verify(number_bits // 2 == rsa_bits(private_key["prime_2"]))
        netius.verify(number_bits == private_key["bits"])

    message = b"Hello World"
    signature = rsa_sign(b"Hello World", private_key)
    result = rsa_verify(signature, private_key)
    result = result.lstrip(b"\0")

    netius.verify(result == message)

def rsa_private(number_bits):
    """
    Generates a new "random" private with the requested number
    of bits as the base for exponents and modulus.

    This method is extremely time consuming in terms of processor
    and should be used carefully to avoid any problem.

    :type number_bits: int
    :param number_bits: The number of bits that are going to be
    used for the generation of the private key.
    :rtype: Dictionary
    :return: The generated private key structure, may then be used
    for processing or written to a file.
    """

    while True:
        prime_1, prime_2 = rsa_primes(number_bits // 2)
        public_exponent, private_exponent = rsa_exponents(prime_1, prime_2, number_bits // 2)
        if private_exponent > 0: break

    modulus = prime_1 * prime_2
    exponent_1 = private_exponent % (prime_1 - 1)
    exponent_2 = private_exponent % (prime_2 - 1)
    coefficient = calc.modinv(prime_2, prime_1)
    bits = rsa_bits(modulus)

    private_key = dict(
        version = 0,
        modulus = modulus,
        public_exponent = public_exponent,
        private_exponent = private_exponent,
        prime_1 = prime_1,
        prime_2 = prime_2,
        exponent_1 = exponent_1,
        exponent_2 = exponent_2,
        coefficient = coefficient,
        bits = bits
    )

    return private_key

def rsa_primes(number_bits):
    """
    Generates two different prime numbers (p and q values)
    and returns them inside a tuple structure.

    The generation is made according to the number of bits
    defined and using a trial and error strategy (expensive).

    :type number_bits: int
    :param number_bits: The number of bits to be used in
    prime generation, this affects security.
    :rtype: Tuple
    :return: A tuple containing the two different prime
    numbers to be returned.
    """

    # calculates the total number of bits for the key as
    # the double of the requested for the primes generation
    total_bits = number_bits * 2

    # constructs the clojure based function that is going to
    # be used as the validator of the primes combination, this
    # used for a trial an error based approach for the generation
    # of the primes to be used in the private key
    def rsa_acceptable(prime_1, prime_2):
        if prime_1 == prime_2: return False
        modulus_bits = rsa_bits(prime_1 * prime_2)
        return modulus_bits == total_bits

    # generates the "first" version of both prime values
    # that is going to serve as the first iteration of
    # each of the values, and start the is odd variable
    prime_1 = calc.prime(number_bits)
    prime_2 = calc.prime(number_bits)
    is_odd = True

    # iterates continuously trying to find a combination
    # of prime numbers that is acceptable and valid
    while True:
        if rsa_acceptable(prime_1, prime_2): break
        if is_odd: prime_1 = calc.prime(number_bits)
        else: prime_2 = calc.prime(number_bits)

    # returns a tuple containing both of the generated
    # primes and returns it to the caller method
    return (prime_1, prime_2)

def rsa_exponents(prime_1, prime_2, number_bits, basic = True):
    """
    Generates both the public and the private exponents for
    the rsa cryptography system taking as base the provided
    prime numbers and the amount of bits for the values.

    :type prime_1: int
    :param prime_1: The first prime number use for rsa.
    :type prime_2: int
    :param prime_2: The second prime number use for rsa.
    :type number_bits: int
    :param number_bits: The number of bits that are going to be
    used for the generation of the values.
    :type basic: bool
    :param basic: If the basic approach to the generation of the
    public exponent should be taken into account.
    :rtype: Tuple
    :return: The tuple containing the generated public and
    private keys (properly tested).
    """

    # calculates the modulus and the phi value for the
    # modulus, as the y are going to be used for calculus
    modulus = prime_1 * prime_2
    phi_modulus = (prime_1 - 1) * (prime_2 - 1)

    # starts by setting the is first flag so that the first iteration
    # of the public exponent generation cycle is taking into account
    # the possible basic flag value for the public exponent
    is_first = True

    # iterates continuously to find a valid public exponent, one
    # that satisfies the relative prime
    while True:
        # make sure e has enough bits so we ensure "wrapping" through
        # modulus (n value) note that if the this is the first attempt
        # to create a public exponent and the basic mode is active the
        # number chosen is the "magic" number (compatibility)
        public_exponent = calc.prime(max(8, number_bits // 2))
        if is_first and basic: public_exponent = 65537; is_first = False

        # checks if the exponent and the modulus are relative primes
        # and also checks if the exponent and the phi modulus are relative
        # primes, for that situation a valid public exponent has been fond
        # and the cycle may be broken
        is_relative = calc.relatively_prime(public_exponent, modulus)
        is_relative_phi = calc.relatively_prime(public_exponent, phi_modulus)
        if is_relative and is_relative_phi: break

    # retrieves the result of the extended euclid greatest common divisor,
    # this value is going to be used as the basis for the calculus of the
    # private exponent for the current operation
    d, l, _e = calc.egcd(public_exponent, phi_modulus)
    private_exponent = l

    # in case the greatest common divisor between both is not one, the values
    # are not relative primes and an exception must be raised
    if not d == 1: raise netius.GeneratorError(
        "The public exponent '%d' and the phi modulus '%d' are not relative primes" %
        (public_exponent, phi_modulus)
    )

    # calculates the inverse modulus for both exponent and in case it's not one
    # an exception is raised about the problem
    inverse_modulus = (public_exponent * private_exponent) % phi_modulus
    if not inverse_modulus == 1: netius.GeneratorError(
        "The public exponent '%d' and private exponent '%d' are not multiplicative inverse modulus of phi modulus '%d'" %
        (public_exponent, private_exponent, phi_modulus)
    )

    # creates the tuple that contains both the public and the private
    # exponent values that may be used for rsa based cryptography
    return (public_exponent, private_exponent)

def rsa_bits(modulus):
    bits = math.log(modulus, 2)
    return calc.ceil_integer(bits)

def rsa_sign(message, private_key):
    message = netius.legacy.bytes(message)
    modulus = private_key["modulus"]
    private_exponent = private_key["private_exponent"]
    return rsa_crypt_s(message, private_exponent, modulus)

def rsa_verify(signature, public_key):
    signature = netius.legacy.bytes(signature)
    modulus = public_key["modulus"]
    public_exponent = public_key["public_exponent"]
    return rsa_crypt_s(signature, public_exponent, modulus)

def rsa_crypt_s(message, exponent, modulus):
    modulus_l = calc.ceil_integer(math.log(modulus, 256))

    message_i = util.bytes_to_integer(message)
    message_crypt = rsa_crypt(message_i, exponent, modulus)
    message_crypt_s = util.integer_to_bytes(message_crypt, modulus_l)

    return message_crypt_s

def rsa_crypt(number, exponent, modulus):
    if not isinstance(number, netius.legacy.INTEGERS):
        raise TypeError("you must pass a long or an int")

    if number > 0 and math.floor(math.log(number, 2)) > math.floor(math.log(modulus, 2)):
        raise OverflowError("the message is too long")

    return pow(number, exponent, modulus)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import struct

import netius

from .base import Middleware

class ProxyMiddleware(Middleware):
    """
    Middleware that implements the PROXY protocol on creation
    of a new connection enabling the passing of information
    from the front-end server to a back-end server using a normal
    TCP connection. This protocol has been development originally
    for the integration of an HAProxy server with back-end servers.

    :see: http://www.haproxy.org/download/1.5/doc/proxy-protocol.txt
    """

    MAX_LENGTH = 118
    """ The maximum length that the base packet may have,
    this is a constant according to PROXY send """

    HEADER_LENGTH_V2 = 16
    """ The length of the header message of the PROXY protocol
    under version 2 """

    HEADER_MAGIC_V2 = b"\x0d\x0a\x0d\x0a\x00\x0d\x0a\x51\x55\x49\x54\x0a"
    """ The magic byte string that starts the PROXY v2 protocol
    header, should be used for runtime verifications """

    TYPE_LOCAL_V2 = 0x0
    TYPE_PROXY_V2 = 0x1

    AF_UNSPEC_v2 = 0x0
    AF_INET_v2 = 0x1
    AF_INET6_v2 = 0x2
    AF_UNIX_v2 = 0x3

    PROTO_UNSPEC_v2 = 0x0
    PROTO_STREAM_v2 = 0x1
    PROTO_DGRAM_v2 = 0x2

    def __init__(self, owner, version = 1):
        Middleware.__init__(self, owner)
        self.version = version

    def start(self):
        Middleware.start(self)
        self.version = netius.conf("PROXY_VERSION", self.version, cast = int)
        self.owner.bind("connection_c", self.on_connection_c)

    def stop(self):
        Middleware.stop(self)
        self.owner.unbind("connection_c", self.on_connection_c)

    def on_connection_c(self, owner, connection):
        if self.version == 1: connection.add_starter(self._proxy_handshake_v1)
        elif self.version == 2: connection.add_starter(self._proxy_handshake_v2)
        else: raise netius.RuntimeError("Invalid PROXY version")

    def _proxy_handshake_v1(self, connection):
        cls = self.__class__

        # verifies if the connection is SSL based if that's the case
        # the safe (reading) mode is enabled
        safe = connection.ssl

        # selects the proper receive method to be used to retrieve bytes
        # from the client side taking into account if the connection is
        # secured with SSL or not, note that the "special" SSL receive method
        # allows one to receive raw information under an SSL socket/connection
        recv = connection._recv_ssl if connection.ssl else connection.recv

        # in case the safe (read) mode is enabled the unit of counting
        # for the receive operation is one (single byte reading) to
        # allow no return of data (required for some environment eg: SSL)
        count = 1 if safe else cls.MAX_LENGTH

        # verifies if there's a previously set PROXY buffer defined
        # for the connection and if that's the case uses it otherwise
        # starts a new empty buffer from scratch
        has_buffer = hasattr(connection, "_proxy_buffer")
        if has_buffer: buffer = connection._proxy_buffer
        else: buffer = bytearray()

        # saves the "newly" created buffer as the PROXY buffer for the
        # current connection (may be used latter)
        connection._proxy_buffer = buffer

        # iterates continuously trying to retrieve the set of data that is
        # required to parse the PROXY protocol header information
        while True:
            # tries to receive the maximum size of data that is required
            # for the handling of the PROXY information
            data = self.owner.exec_safe(connection, recv, count)

            # in case the received data represents that of a closed connection
            # the connection is closed and the control flow returned
            if data == b"": connection.close(); return

            # in case the received value is false, that indicates that the
            # execution has failed due to an exception (expected or unexpected)
            if data == False: return

            # updates the "initial" buffer length taking into account
            # the current buffer and then appends the new data to it
            buffer_l = len(buffer)
            buffer += data

            # verifies the end of line sequence is present in the buffer,
            # if that's the case we've reached a positive state
            is_ready = b"\r\n" in buffer

            # in case the ready state has been reached, the complete set of
            # data is ready to be parsed and the loop is stopped
            if is_ready: break

        # removes the PROXY buffer reference from the connection as
        # its no longer going to be used
        del connection._proxy_buffer

        # determines the index/position of the end sequence in the
        # buffer and then uses it to calculate the base for the
        # calculus of the extra information in the data buffer
        buffer_i = buffer.index(b"\r\n")
        data_b = buffer_i - buffer_l + 2

        # extracts the line for parsing and the extra data value (to
        # be restored to connection) using the data base and the data
        line = buffer[:buffer_i]
        extra = data[data_b:]

        # in case there's valid extra data to be restored to the connection
        # performs the operation, effectively restoring it for latter
        # receiving operations (just like adding it back to the socket)
        if extra: connection.restore(extra)

        # forces the "conversion" of the line into a string so that it may
        # be properly split into its components, note that first the value
        # is converted into a byte string and then into a proper string
        line = bytes(line)
        line = netius.legacy.str(line)

        # splits the line of the protocol around its components and uses them
        # to change the current connection information (as expected)
        header, protocol, source, destination, source_p, destination_p = line.split(" ")

        # prints a debug message about the PROXY header received, so that runtime
        # debugging is possible (and expected for this is a sensible part)
        self.owner.debug(
            "Received header %s %s %s:%s => %s:%s" %
            (header, protocol, source, source_p, destination, destination_p)
        )

        # re-constructs the source address from the provided information, this is
        # the major and most important fix to be done
        connection.address = (source, int(source_p))

        # runs the end starter operation, indicating to the connection that
        # the PROXY header has been properly parsed
        connection.end_starter()

    def _proxy_handshake_v2(self, connection):
        import netius.common

        cls = self.__class__

        # verifies if there's a previously set PROXY buffer defined
        # for the connection and if that's the case uses it otherwise
        # starts a new empty buffer from scratch
        has_buffer = hasattr(connection, "_proxy_buffer")
        if has_buffer: buffer = connection._proxy_buffer
        else: buffer = bytearray()

        # saves the "newly" created buffer as the PROXY buffer for the
        # current connection (may be used latter)
        connection._proxy_buffer = buffer

        # verifies if a PROXY header was already parsed from the current connection
        # and if that was not the case runs its parsing
        header = connection._proxy_header if hasattr(connection, "_proxy_header") else None
        if not header:
            # tries to read the PROXY v2 header bytes to be able to parse
            # the body parts taking that into account
            header = self._read_safe(connection, buffer, cls.HEADER_LENGTH_V2)
            if not header: return

            # updates the reference to the proxy header in the connection
            # and clears the buffer as it's now going to be used to load
            # the data from the body part
            connection._proxy_header = header
            buffer[:] = b""

        # unpacks the PROXY v2 header into its components, notice that some of them
        # contain multiple values on higher and lower bits
        magic, version_type, address_protocol, body_size = struct.unpack("!12sBBH", header)

        # unpacks both the version (of the protocol) and the type (of message) by
        # unpacking the higher and the lower bits
        version = version_type >> 4
        type = version_type & 0x0f

        # unpacks the type of address to be communicated and the protocol family
        address = address_protocol >> 4
        protocol = address_protocol & 0x0f

        # runs a series of assertions on some of the basic promises of the protocol
        # (if they failed connection will be dropped)
        netius.verify(magic == cls.HEADER_MAGIC_V2)
        netius.verify(version == 2)

        # reads the body part of the PROXY message taking into account the advertised
        # size of the body (from header component)
        body = self._read_safe(connection, buffer, body_size)
        if not body: return

        if address == cls.AF_INET_v2:
            source, destination, source_p, destination_p = struct.unpack("!IIHH", body)
            source = netius.common.addr_to_ip4(source)
            destination = netius.common.addr_to_ip4(destination)
        elif address == cls.AF_INET6_v2:
            source_high,\
            source_low,\
            destination_high,\
            destination_low,\
            source_p,\
            destination_p = struct.unpack("!QQQQHH", body)
            source = (source_high << 64) + source_low
            destination = (destination_high << 64) + destination_low
            source = netius.common.addr_to_ip6(source)
            destination = netius.common.addr_to_ip6(destination)
        else:
            raise netius.RuntimeError("Unsupported or invalid PROXY header")

        # removes the PROXY buffer and header references from the connection
        # as they are no longer going to be used
        del connection._proxy_buffer
        del connection._proxy_header

        # prints a debug message about the PROXY header received, so that runtime
        # debugging is possible (and expected for this is a sensible part)
        self.owner.debug(
            "Received header v2 %d %s:%s => %s:%s" %
            (protocol, source, source_p, destination, destination_p)
        )

        # re-constructs the source address from the provided information, this is
        # the major and most important fix to be done
        connection.address = (source, int(source_p))

        # runs the end starter operation, indicating to the connection that
        # the PROXY header has been properly parsed
        connection.end_starter()

    def _read_safe(self, connection, buffer, count):
        """
        Reads a certain amount of data from a non blocking connection,
        in case the're a blocking operation then the error is raised
        and caught by the upper layers.

        This method also assumes that the buffer is stored on an abstract
        layer that can be used in the resume operation.

        :type connection: Connection
        :param connection: The connection from which the data is going
        to be read.
        :type buffer: bytearray
        :param bytearray: The byte array where the data is going to be store
        waiting for the processing.
        :type count: int
        :param count: The number of bytes that are going to be read from
        the target connection.
        :rtype: String
        :return: The bytes that were read from the connection or in alternative
        an invalid value meaning that the connection should be dropped.
        """

        cls = self.__class__

        # selects the proper receive method to be used to retrieve bytes
        # from the client side taking into account if the connection is
        # secured with SSL or not, note that the "special" SSL receive method
        # allows one to receive raw information under an SSL socket/connection
        recv = connection._recv_ssl if connection.ssl else connection.recv

        # iterates continuously trying to retrieve the set of data that is
        # required to parse the PROXY protocol header information
        while True:
            # determines the number of pending bytes in remaining to be read
            # in the buffer and if that's less or equal to zero breaks the
            # current loop (nothing pending to be read)
            pending = count - len(buffer)
            if pending <= 0: break

            # tries to receive the maximum size of data that is required
            # for the handling of the PROXY information
            data = self.owner.exec_safe(connection, recv, pending)

            # in case the received data represents that of a closed connection
            # the connection is closed and the control flow returned
            if data == b"": connection.close(); return None

            # in case the received value is false, that indicates that the
            # execution has failed due to an exception (expected or unexpected)
            if data == False: return None

            # adds the newly read data to the current buffer
            buffer += data

        # returns the valid partial value of the buffer as requested by
        # the call to this method, in normal circumstances the buffer
        # should only contain the requested amount of data
        return bytes(buffer[:count])



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

class Observable(object):
    """
    The base class that implements the observable
    patter allowing the object to handle a series of
    event in a dynamic fashion.

    This class should be friendly to multiple inheritance
    and should avoid variable naming collision.
    """

    def __init__(self, *args, **kwargs):
        self.events = {}

    def build(self):
        pass

    def destroy(self):
        self.unbind_all()

    def bind(self, name, method, oneshot = False):
        if oneshot: method.oneshot = oneshot
        methods = self.events.get(name, [])
        methods.append(method)
        self.events[name] = methods

    def unbind(self, name, method = None):
        methods = self.events.get(name, None)
        if not methods: return
        if method: methods.remove(method)
        else: del methods[:]

    def unbind_all(self):
        if not hasattr(self, "events"): return
        for methods in self.events.values(): del methods[:]
        self.events.clear()

    def trigger(self, name, *args, **kwargs):
        methods = self.events.get(name, None)
        if not methods: return
        oneshots = None
        for method in methods:
            method(*args, **kwargs)
            if not hasattr(method, "oneshot"): continue
            if not method.oneshot: continue
            oneshots = [] if oneshots == None else oneshots
            oneshots.append(method)
        if not oneshots: return
        for oneshot in oneshots: self.unbind(name, oneshot)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os

import netius

class TLSContextDict(dict):

    def __init__(self, owner, domains, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.owner = owner
        self.load(domains)

    def load(self, domains):
        secure = self.owner.get_env("SSL_SECURE", True, cast = bool)
        for domain in domains:
            if not self.has_definition(domain): continue
            cer_path = self.cer_path(domain)
            key_path = self.key_path(domain)
            values = dict(cer_file = cer_path, key_file = key_path)
            context = self.owner._ssl_ctx(values, secure = secure)
            self[domain] = (context, values)

    def has_definition(self, domain):
        cer_path = self.cer_path(domain)
        key_path = self.key_path(domain)
        if not os.path.exists(cer_path): return False
        if not os.path.exists(key_path): return False
        return True

    def cer_path(self, domain):
        raise netius.NotImplemented("Missing implementation")

    def key_path(self, domain):
        raise netius.NotImplemented("Missing implementation")

class LetsEncryptDict(TLSContextDict):

    def __init__(self, owner, domains, *args, **kwargs):
        self.letse_path = kwargs.get("letse_path", "/data/letsencrypt/etc/live")
        TLSContextDict.__init__(self, owner, domains, *args, **kwargs)

    def cer_path(self, domain):
        domain_path = os.path.join(self.letse_path, domain)
        return os.path.join(domain_path, "fullchain.pem")

    def key_path(self, domain):
        domain_path = os.path.join(self.letse_path, domain)
        return os.path.join(domain_path, "privkey.pem")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius

from . import base

class MemoryAdapter(base.BaseAdapter):

    def __init__(self):
        base.BaseAdapter.__init__(self)
        self.map = dict()
        self.owners = dict()

    def set(self, value, owner = "nobody"):
        map_o = self._ensure(owner)
        key = self.generate()
        item = dict(value = value, owner = owner)
        self.map[key] = item
        map_o[key] = item
        return key

    def get_file(self, key, mode = "rb"):
        if not key in self.map: netius.NetiusError("Key not found")
        item = self.map[key]
        value = item["value"]
        file = netius.legacy.StringIO(value)
        close = self._build_close(file, key)
        file._close = file.close
        file.close = close
        return file

    def delete(self, key, owner = "nobody"):
        item = self.map[key]
        owner = item["owner"]
        map_o = self._ensure(owner)
        del self.map[key]
        del map_o[key]

    def append(self, key, value):
        item = self.map[key]
        _value = item["value"]
        _value += value
        item["value"] = _value

    def truncate(self, key, count):
        item = self.map[key]
        _value = item["value"]
        offset = count * -1
        _value = _value[:offset]
        item["value"] = _value

    def size(self, key):
        item = self.map[key]
        _value = item["value"]
        return len(_value)

    def count(self, owner = None):
        map = self._ensure(owner) if owner else self.map
        return len(map)

    def list(self, owner = None):
        map = self._ensure(owner) if owner else self.map
        return map.keys()

    def _ensure(self, owner):
        map = self.owners.get(owner, {})
        self.owners[owner] = map
        return map

    def _build_close(self, file, key):

        def close():
            value = file.getvalue()
            item = self.map[key]
            item["value"] = value
            file._close()

        return close



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import re

from . import errors

FIRST_CAP_REGEX = re.compile("(.)([A-Z][a-z]+)")
""" Regular expression that ensures that the first
token of each camel string is properly capitalized """

ALL_CAP_REGEX = re.compile("([a-z0-9])([A-Z])")
""" The generalized transition from lower case to
upper case letter regex that will provide a way of
putting the underscore in the middle of the transition """

def camel_to_underscore(camel, separator = "_"):
    """
    Converts the provided camel cased based value into
    a normalized underscore based string.

    This is useful as most of the python string standards
    are compliant with the underscore strategy.

    :type camel: String
    :param camel: The camel cased string that is going to be
    converted into an underscore based string.
    :type separator: String
    :param separator: The separator token that is going to
    be used in the camel to underscore conversion.
    :rtype: String
    :return: The underscore based string resulting from the
    conversion of the provided camel cased one.
    """

    value = FIRST_CAP_REGEX.sub(r"\1" + separator + r"\2", camel)
    value = ALL_CAP_REGEX.sub(r"\1" + separator + r"\2", value)
    value = value.lower()
    return value

def verify(condition, message = None, exception = None):
    """
    Ensures that the requested condition returns a valid value
    and if that's no the case an exception raised breaking the
    current execution logic.

    :type condition: bool
    :param condition: The condition to be evaluated and that may
    trigger an exception raising.
    :type message: String
    :param message: The message to be used in the building of the
    exception that is going to be raised in case of condition failure.
    :type exception: Class
    :param exception: The exception class that is going to be used
    to build the exception to be raised in case the condition
    verification operation fails.
    """

    if condition: return
    exception = exception or errors.AssertionError
    raise exception(message or "Assertion Error")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius

from . import common

TASK_WORK = 10

class TaskThread(common.Thread):

    def execute(self, work):
        type = work[0]
        if not type == TASK_WORK: netius.NotImplemented(
            "Cannot execute type '%d'" % type
        )

        callable, args, kwargs, callback = work[1:]
        result = callable(*args, **kwargs)
        if callback: callback(result)

class TaskPool(common.EventPool):

    def __init__(self, base = TaskThread, count = 10):
        common.EventPool.__init__(self, base = base, count = count)

    def execute(self, callable, args = [], kwargs = {}, callback = None):
        work = (TASK_WORK, callable, args, kwargs, callback)
        self.push(work)
